"""Async SQLAlchemy connection helpers."""

from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
import shlex
from typing import Optional
from urllib.parse import parse_qsl, urlencode, urlsplit, urlunsplit

# Import duckdb_engine to register the dialect with SQLAlchemy
try:
    import duckdb_engine  # noqa: F401
except ImportError:
    pass

try:
    from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
    from sqlalchemy.ext.asyncio.engine import AsyncConnection
except ImportError as exc:
    raise ImportError(
        "Async support requires SQLAlchemy 2.0+ with async extensions. "
        "Install with: pip install 'SQLAlchemy>=2.0'"
    ) from exc

from sqlalchemy import text

from ..config import EngineConfig
from ..engine.dialects import DialectSpec, get_dialect


def _extract_postgres_server_settings(dsn: str) -> tuple[str, dict[str, str]]:
    """Convert Postgres DSN ``options`` into asyncpg server settings.

    asyncpg does not support the ``options`` keyword argument that psycopg does.
    SQLAlchemy forwards query parameters from the DSN (e.g. ?options=-csearch_path=foo)
    directly to asyncpg, which raises ``TypeError``. We translate any ``-cKEY=VALUE``
    tokens into asyncpg ``server_settings`` entries and drop the ``options`` query param.
    """

    split = urlsplit(dsn)
    scheme = split.scheme.split("+")[0]
    if scheme != "postgresql":
        return dsn, {}

    query_items = parse_qsl(split.query, keep_blank_values=True)
    if not query_items:
        return dsn, {}

    server_settings: dict[str, str] = {}
    filtered_query: list[tuple[str, str]] = []

    for key, value in query_items:
        if key != "options" or not value:
            filtered_query.append((key, value))
            continue

        for token in shlex.split(value):
            if not token.startswith("-c"):
                continue
            setting = token[2:]
            if "=" not in setting:
                continue
            name, setting_value = setting.split("=", 1)
            name = name.strip()
            if not name:
                continue
            server_settings[name] = setting_value.strip()

    if not server_settings:
        return dsn, {}

    new_query = urlencode(filtered_query, doseq=True)
    normalized = urlunsplit((split.scheme, split.netloc, split.path, new_query, split.fragment))
    return normalized, server_settings


class AsyncConnectionManager:
    """Creates and caches async SQLAlchemy engines for Moltres sessions."""

    def __init__(self, config: EngineConfig):
        self.config = config
        self._engine: AsyncEngine | None = None
        self._session: object | None = None  # SQLAlchemy AsyncSession
        self._active_transaction: Optional[AsyncConnection] = None
        self._savepoint_stack: list[str] = []
        self._transaction_metadata: Optional[dict[str, object]] = None

    def _create_engine(self) -> AsyncEngine:
        """Create an async SQLAlchemy engine.

        Args:
            config: Engine configuration

        Returns:
            AsyncEngine instance

        Raises:
            ValueError: If DSN doesn't support async (missing +asyncpg, +aiomysql, etc.)
        """
        # If a session is provided, extract engine from it
        if self.config.session is not None:
            session = self.config.session
            # Check if it's a SQLAlchemy AsyncSession or SQLModel AsyncSession
            bind = None
            # For AsyncSession, prefer .bind over .get_bind() because get_bind() returns sync Engine
            if hasattr(session, "bind"):
                bind = session.bind
            elif hasattr(session, "get_bind"):
                # SQLAlchemy 2.0 style - but get_bind() might return sync Engine for async sessions
                try:
                    bind = session.get_bind()
                    # If get_bind() returned a sync Engine, it's not what we want
                    if not isinstance(bind, AsyncEngine):
                        bind = None
                except (TypeError, AttributeError):
                    # get_bind() might require arguments
                    pass
            if bind is None:
                # Try to get from sessionmaker or other attributes
                if hasattr(session, "maker") and hasattr(session.maker, "bind"):
                    bind = session.maker.bind
            if bind is None:
                raise TypeError(
                    "session must be a SQLAlchemy AsyncSession or SQLModel AsyncSession instance "
                    f"with a bind (engine) attached. Got: {type(session).__name__}"
                )
            if not isinstance(bind, AsyncEngine):
                raise TypeError(
                    "Session's bind must be an AsyncEngine, not a synchronous Engine. "
                    f"Got: {type(bind).__name__}. Use connect() for sync sessions."
                )
            self._session = session
            return bind

        # If an engine is provided in config, use it directly
        if self.config.engine is not None:
            if not isinstance(self.config.engine, AsyncEngine):
                raise TypeError("engine must be a SQLAlchemy AsyncEngine instance")
            return self.config.engine

        # Otherwise, create a new engine from DSN
        if self.config.dsn is None:
            raise ValueError(
                "Either 'dsn', 'engine', or 'session' must be provided in EngineConfig"
            )

        dsn = self.config.dsn

        # Check if DSN already has async driver specified
        dsn_parts = dsn.split("://", 1)
        if len(dsn_parts) < 2:
            raise ValueError(f"Invalid DSN format: {dsn}")

        scheme = dsn_parts[0]
        # Check if scheme already has an async driver
        has_async_driver = "+" in scheme and any(
            driver in scheme for driver in ["asyncpg", "aiomysql", "aiosqlite"]
        )

        if not has_async_driver:
            # Auto-detect and add async driver based on database type
            # Handle schemes with sync drivers (e.g., mysql+pymysql -> mysql+aiomysql)
            base_scheme = scheme.split("+")[
                0
            ]  # Get base scheme (e.g., "mysql" from "mysql+pymysql")

            if base_scheme == "postgresql":
                # Replace any existing driver with asyncpg
                if "+" in scheme:
                    dsn = dsn.replace(scheme, "postgresql+asyncpg", 1)
                else:
                    dsn = dsn.replace("postgresql://", "postgresql+asyncpg://", 1)
            elif base_scheme in ("mysql", "mariadb"):
                # Replace any existing driver with aiomysql
                if "+" in scheme:
                    dsn = dsn.replace(scheme, f"{base_scheme}+aiomysql", 1)
                else:
                    dsn = dsn.replace("mysql://", "mysql+aiomysql://", 1).replace(
                        "mariadb://", "mariadb+aiomysql://", 1
                    )
            elif base_scheme == "sqlite":
                # Replace any existing driver with aiosqlite
                if "+" in scheme:
                    dsn = dsn.replace(scheme, "sqlite+aiosqlite", 1)
                else:
                    dsn = dsn.replace("sqlite:///", "sqlite+aiosqlite:///", 1)
            else:
                raise ValueError(
                    f"DSN '{dsn}' does not specify an async driver. "
                    "Use format like 'postgresql+asyncpg://...' or 'mysql+aiomysql://...'"
                )

        # Refresh scheme after any driver normalization
        scheme = dsn.split("://", 1)[0]

        kwargs: dict[str, object] = {"echo": self.config.echo}
        if self.config.pool_size is not None:
            kwargs["pool_size"] = self.config.pool_size
        if self.config.max_overflow is not None:
            kwargs["max_overflow"] = self.config.max_overflow
        if self.config.pool_timeout is not None:
            kwargs["pool_timeout"] = self.config.pool_timeout
        if self.config.pool_recycle is not None:
            kwargs["pool_recycle"] = self.config.pool_recycle
        if self.config.pool_pre_ping:
            kwargs["pool_pre_ping"] = self.config.pool_pre_ping

        if "postgresql+asyncpg" in scheme:
            normalized_dsn, server_settings = _extract_postgres_server_settings(dsn)
            if server_settings:
                dsn = normalized_dsn
                connect_args_obj = kwargs.setdefault("connect_args", {})
                if not isinstance(connect_args_obj, dict):
                    raise TypeError("connect_args must be a dict")
                server_settings_container = connect_args_obj.setdefault("server_settings", {})
                if not isinstance(server_settings_container, dict):
                    raise TypeError("server_settings connect arg must be a dict")
                server_settings_container.update(server_settings)

        return create_async_engine(dsn, **kwargs)

    @property
    def engine(self) -> AsyncEngine:
        """Get or create the async engine."""
        if self._engine is None:
            self._engine = self._create_engine()
        return self._engine

    @asynccontextmanager
    async def connect(
        self, transaction: Optional[AsyncConnection] = None
    ) -> AsyncIterator[AsyncConnection]:
        """Get an async database connection.

        Args:
            transaction: If provided, use this transaction connection instead of creating a new one.
                        This allows operations to share a transaction.
                        If None and an active transaction exists, uses the active transaction.

        Yields:
            AsyncConnection instance
        """
        if transaction is not None:
            # Use the provided transaction connection
            yield transaction
        elif self._active_transaction is not None:
            # Use the active transaction connection automatically
            yield self._active_transaction
        elif self._session is not None:
            # Use the session's connection
            # SQLAlchemy async sessions have a connection() method that returns a coroutine
            if hasattr(self._session, "connection"):
                # Get connection from session (async)
                connection = await self._session.connection()
                yield connection
            else:
                # Fallback: use session's bind to create a connection
                async with self.engine.begin() as connection:
                    yield connection
        else:
            # Create a new connection with auto-commit (default behavior)
            async with self.engine.begin() as connection:
                yield connection

    async def begin_transaction(
        self,
        savepoint: bool = False,
        readonly: bool = False,
        isolation_level: Optional[str] = None,
        timeout: Optional[float] = None,
    ) -> AsyncConnection:
        """Begin a new transaction and return the connection.

        Args:
            savepoint: If True and a transaction is already active, create a savepoint instead.
            readonly: If True, set transaction to read-only mode.
            isolation_level: Optional isolation level (READ UNCOMMITTED, READ COMMITTED,
                           REPEATABLE READ, SERIALIZABLE).
            timeout: Optional transaction timeout in seconds.

        Returns:
            AsyncConnection that is part of a transaction (not auto-committed)

        Raises:
            RuntimeError: If savepoint=False and a transaction is already active.
            ValueError: If isolation level or readonly is requested but not supported by dialect.
        """
        if self._active_transaction is not None:
            if savepoint:
                # Create a savepoint instead of a new transaction
                savepoint_name = self._generate_savepoint_name()
                return await self.create_savepoint(self._active_transaction, savepoint_name)
            else:
                raise RuntimeError(
                    "Transaction already active. Use savepoint=True for nested transactions."
                )

        # Get dialect for feature checking
        dialect_name = self.engine.dialect.name
        try:
            dialect_spec = get_dialect(dialect_name)
        except ValueError:
            # Unknown dialect, use conservative defaults
            dialect_spec = DialectSpec(name=dialect_name)

        self._active_transaction = await self.engine.connect()
        self._savepoint_stack = []
        self._transaction_metadata = {
            "readonly": readonly,
            "isolation_level": isolation_level,
            "timeout": timeout,
        }

        # Set isolation level if specified
        if isolation_level:
            if not dialect_spec.supports_isolation_levels:
                await self._active_transaction.close()
                self._active_transaction = None
                raise ValueError(
                    f"Dialect '{dialect_name}' does not support isolation levels. "
                    "SQLite only supports SERIALIZABLE and READ UNCOMMITTED via PRAGMA."
                )
            await self._set_isolation_level(self._active_transaction, isolation_level)

        # Set read-only mode if specified
        if readonly:
            if not dialect_spec.supports_read_only_transactions:
                await self._active_transaction.close()
                self._active_transaction = None
                raise ValueError(
                    f"Dialect '{dialect_name}' does not support read-only transactions."
                )
            await self._set_readonly(self._active_transaction, True)

        # Set timeout if specified
        if timeout:
            await self._set_timeout(self._active_transaction, timeout, dialect_name)

        await self._active_transaction.begin()
        return self._active_transaction

    def _generate_savepoint_name(self) -> str:
        """Generate a unique savepoint name."""
        return f"sp_{len(self._savepoint_stack)}"

    async def _set_isolation_level(self, connection: AsyncConnection, isolation_level: str) -> None:
        """Set transaction isolation level."""
        # Normalize isolation level names
        level_map = {
            "READ UNCOMMITTED": "READ UNCOMMITTED",
            "READ COMMITTED": "READ COMMITTED",
            "REPEATABLE READ": "REPEATABLE READ",
            "SERIALIZABLE": "SERIALIZABLE",
        }
        normalized = level_map.get(isolation_level.upper())
        if not normalized:
            raise ValueError(
                f"Invalid isolation level '{isolation_level}'. "
                "Must be one of: READ UNCOMMITTED, READ COMMITTED, REPEATABLE READ, SERIALIZABLE"
            )

        stmt = text(f"SET TRANSACTION ISOLATION LEVEL {normalized}")
        await connection.execute(stmt)

    async def _set_readonly(self, connection: AsyncConnection, readonly: bool) -> None:
        """Set transaction to read-only mode."""
        mode = "READ ONLY" if readonly else "READ WRITE"
        stmt = text(f"SET TRANSACTION {mode}")
        await connection.execute(stmt)

    async def _set_timeout(
        self, connection: AsyncConnection, timeout: float, dialect_name: str
    ) -> None:
        """Set transaction timeout (database-specific)."""
        # PostgreSQL uses statement_timeout (in milliseconds)
        if "postgresql" in dialect_name:
            stmt = text(f"SET statement_timeout = {int(timeout * 1000)}")
            await connection.execute(stmt)
        # MySQL uses innodb_lock_wait_timeout (in seconds)
        elif "mysql" in dialect_name:
            stmt = text(f"SET innodb_lock_wait_timeout = {int(timeout)}")
            await connection.execute(stmt)
        # SQLite doesn't support transaction timeouts directly
        # Other databases may need specific implementations

    async def create_savepoint(self, connection: AsyncConnection, name: str) -> AsyncConnection:
        """Create a savepoint in the current transaction.

        Args:
            connection: The transaction connection
            name: Savepoint name

        Returns:
            The same connection (for compatibility)

        Raises:
            RuntimeError: If no transaction is active or connection doesn't match active transaction.
        """
        if connection is not self._active_transaction:
            raise RuntimeError("Connection is not the active transaction")
        if not self._active_transaction:
            raise RuntimeError("No active transaction")

        # Get dialect for feature checking
        dialect_name = self.engine.dialect.name
        try:
            dialect_spec = get_dialect(dialect_name)
        except ValueError:
            dialect_spec = DialectSpec(name=dialect_name)

        if not dialect_spec.supports_savepoints:
            raise ValueError(f"Dialect '{dialect_name}' does not support savepoints.")

        stmt = text(f"SAVEPOINT {name}")
        await connection.execute(stmt)
        self._savepoint_stack.append(name)
        return connection

    async def rollback_to_savepoint(self, connection: AsyncConnection, name: str) -> None:
        """Rollback to a specific savepoint.

        Args:
            connection: The transaction connection
            name: Savepoint name to rollback to

        Raises:
            RuntimeError: If no transaction is active, connection doesn't match, or savepoint not found.
        """
        if connection is not self._active_transaction:
            raise RuntimeError("Connection is not the active transaction")
        if not self._active_transaction:
            raise RuntimeError("No active transaction")
        if name not in self._savepoint_stack:
            raise RuntimeError(f"Savepoint '{name}' not found in current transaction")

        stmt = text(f"ROLLBACK TO SAVEPOINT {name}")
        await connection.execute(stmt)

        # Remove all savepoints after the one we're rolling back to
        index = self._savepoint_stack.index(name)
        self._savepoint_stack = self._savepoint_stack[: index + 1]

    async def release_savepoint(self, connection: AsyncConnection, name: str) -> None:
        """Release a savepoint.

        Args:
            connection: The transaction connection
            name: Savepoint name to release

        Raises:
            RuntimeError: If no transaction is active, connection doesn't match, or savepoint not found.
        """
        if connection is not self._active_transaction:
            raise RuntimeError("Connection is not the active transaction")
        if not self._active_transaction:
            raise RuntimeError("No active transaction")
        if name not in self._savepoint_stack:
            raise RuntimeError(f"Savepoint '{name}' not found in current transaction")

        stmt = text(f"RELEASE SAVEPOINT {name}")
        await connection.execute(stmt)
        self._savepoint_stack.remove(name)

    async def commit_transaction(self, connection: AsyncConnection) -> None:
        """Commit a transaction.

        Args:
            connection: The transaction connection to commit
        """
        if connection is not self._active_transaction:
            raise RuntimeError("Connection is not the active transaction")
        try:
            await connection.commit()
        finally:
            # Always close connection, even if commit fails
            await connection.close()
            self._active_transaction = None
            self._savepoint_stack = []
            self._transaction_metadata = None

    async def rollback_transaction(self, connection: AsyncConnection) -> None:
        """Rollback a transaction.

        Args:
            connection: The transaction connection to rollback
        """
        if connection is not self._active_transaction:
            raise RuntimeError("Connection is not the active transaction")
        try:
            await connection.rollback()
        finally:
            # Always close connection, even if rollback fails
            await connection.close()
            self._active_transaction = None
            self._savepoint_stack = []
            self._transaction_metadata = None

    @property
    def active_transaction(self) -> Optional[AsyncConnection]:
        """Get the active transaction connection if one exists."""
        return self._active_transaction

    @property
    def transaction_metadata(self) -> Optional[dict[str, object]]:
        """Get transaction metadata if a transaction is active."""
        return self._transaction_metadata

    @property
    def savepoint_stack(self) -> list[str]:
        """Get the current savepoint stack."""
        return self._savepoint_stack.copy()

    async def close(self) -> None:
        """Close the engine and all connections."""
        if self._engine is not None:
            await self._engine.dispose()
            self._engine = None



"""Transaction decorator for automatic transaction management.

This module provides a decorator that automatically wraps functions in database transactions,
simplifying transaction management for common use cases.
"""

from __future__ import annotations

import functools
import logging
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar, Union, overload

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ..table.table import Database
    from ..table.async_table import AsyncDatabase

T = TypeVar("T")
F = TypeVar("F", bound=Callable[..., Any])


@overload
def transaction(
    database: "Database",
    *,
    savepoint: bool = False,
    readonly: bool = False,
    isolation_level: Optional[str] = None,
    timeout: Optional[float] = None,
) -> Callable[[F], F]: ...


@overload
def transaction(
    database: "AsyncDatabase",
    *,
    savepoint: bool = False,
    readonly: bool = False,
    isolation_level: Optional[str] = None,
    timeout: Optional[float] = None,
) -> Callable[[F], F]: ...


@overload
def transaction(
    *,
    savepoint: bool = False,
    readonly: bool = False,
    isolation_level: Optional[str] = None,
    timeout: Optional[float] = None,
) -> Callable[[F], F]: ...


def transaction(
    database: Optional[Union["Database", "AsyncDatabase"]] = None,
    *,
    savepoint: bool = False,
    readonly: bool = False,
    isolation_level: Optional[str] = None,
    timeout: Optional[float] = None,
) -> Union[Callable[[F], F], Callable[[Callable[..., Any]], Callable[..., Any]]]:
    """Decorator to automatically wrap a function in a database transaction.

    The decorator can be used in two ways:

    1. With a database instance (recommended for class methods or functions with db parameter):
       ```python
       db = connect("sqlite:///example.db")

       @transaction(db)
       def create_user(name: str):
           Records(_data=[{"name": name}], _database=db).insert_into("users")
       ```

    2. Without a database instance (database is expected as first parameter or keyword arg):
       ```python
       @transaction
       def create_user(db: Database, name: str):
           Records(_data=[{"name": name}], _database=db).insert_into("users")
       ```

    For async functions, the decorator will automatically detect and handle async transactions:
       ```python
       @transaction(async_db)
       async def create_user_async(name: str):
           await AsyncRecords(_data=[{"name": name}], _database=async_db).insert_into("users")
       ```

    Args:
        database: Optional database instance. If provided, this database will be used for
                 all transactions. If not provided, the database is expected as the first
                 parameter or a 'db' keyword argument in the decorated function.
        savepoint: Use savepoint for nested transactions (default: False)
        readonly: Create read-only transaction (default: False)
        isolation_level: Set transaction isolation level (e.g., "SERIALIZABLE", "READ COMMITTED")
        timeout: Transaction timeout in seconds (default: None)

    Returns:
        Decorated function that automatically runs in a transaction

    Example:
        >>> from moltres import connect
        >>> from moltres.utils.transaction_decorator import transaction
        >>> from moltres.io.records import Records
        >>>
        >>> db = connect("sqlite:///:memory:")
        >>> _ = db.create_table("users", [column("id", "INTEGER"), column("name", "TEXT")]).collect()
        >>>
        >>> # Method 1: Provide database instance
        >>> @transaction(db)
        ... def add_user(name: str):
        ...     Records(_data=[{"id": 1, "name": name}], _database=db).insert_into("users")
        >>>
        >>> # Method 2: Database as parameter
        >>> @transaction
        ... def add_user_with_db(db: Database, name: str):
        ...     Records(_data=[{"id": 2, "name": name}], _database=db).insert_into("users")
        >>>
        >>> add_user("Alice")
        >>> add_user_with_db(db, "Bob")
        >>> results = db.table("users").select().collect()
        >>> assert len(results) == 2
    """
    # If database is None, this is the decorator being called without arguments
    # We need to return a decorator that expects the function
    if database is None:
        # Called as @transaction
        def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
            return _wrap_with_transaction(
                func,
                database=None,
                savepoint=savepoint,
                readonly=readonly,
                isolation_level=isolation_level,
                timeout=timeout,
            )

        return decorator

    # If database is provided, check if it's actually a function (called as @transaction(db))
    # or if it's the database instance (called as @transaction(db))
    if callable(database):
        # Called as @transaction with function (database is actually the function)
        func = database
        return _wrap_with_transaction(
            func,
            database=None,
            savepoint=savepoint,
            readonly=readonly,
            isolation_level=isolation_level,
            timeout=timeout,
        )

    # Called as @transaction(db) - database is the Database instance
    def db_decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        return _wrap_with_transaction(
            func,
            database=database,
            savepoint=savepoint,
            readonly=readonly,
            isolation_level=isolation_level,
            timeout=timeout,
        )

    return db_decorator


def _wrap_with_transaction(
    func: Callable[..., Any],
    database: Optional[Union["Database", "AsyncDatabase"]] = None,
    savepoint: bool = False,
    readonly: bool = False,
    isolation_level: Optional[str] = None,
    timeout: Optional[float] = None,
) -> Callable[..., Any]:
    """Wrap a function to run in a transaction."""
    import inspect

    is_async = inspect.iscoroutinefunction(func)

    if is_async:
        # Async function

        @functools.wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            # Determine which database to use
            db: Any = database
            if db is None:
                # Try to find database in args or kwargs
                for arg in args:
                    from ..table.async_table import AsyncDatabase

                    if isinstance(arg, AsyncDatabase):
                        db = arg
                        break
                if db is None:
                    db = kwargs.get("db")
                    if db is None:
                        raise ValueError(
                            "Database instance not found. Provide database as first argument "
                            "or 'db' keyword argument, or pass database to @transaction decorator."
                        )

            # Run function in transaction
            async with db.transaction(
                savepoint=savepoint,
                readonly=readonly,
                isolation_level=isolation_level,
                timeout=timeout,
            ):
                return await func(*args, **kwargs)

        return async_wrapper
    else:
        # Sync function

        @functools.wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            # Determine which database to use
            db: Any = database
            if db is None:
                # Try to find database in args or kwargs
                for arg in args:
                    from ..table.table import Database

                    if isinstance(arg, Database):
                        db = arg
                        break
                if db is None:
                    db = kwargs.get("db")
                    if db is None:
                        raise ValueError(
                            "Database instance not found. Provide database as first argument "
                            "or 'db' keyword argument, or pass database to @transaction decorator."
                        )

            # Run function in transaction
            with db.transaction(
                savepoint=savepoint,
                readonly=readonly,
                isolation_level=isolation_level,
                timeout=timeout,
            ):
                return func(*args, **kwargs)

        return sync_wrapper



"""Database query execution manager.

This module handles query execution operations including plan execution, SQL execution, and explain.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Type

if TYPE_CHECKING:
    from ..engine.execution import QueryResult
    from ..logical.plan import LogicalPlan
    from .table import Database


class DatabaseQueryExecutor:
    """Handles query execution operations for Database."""

    def __init__(self, database: "Database"):
        """Initialize query executor with a Database.

        Args:
            database: The Database instance to execute queries on
        """
        self._db = database

    def execute_plan(self, plan: "LogicalPlan", model: Optional[Type[Any]] = None) -> "QueryResult":
        """Execute a logical plan and return results.

        Args:
            plan: Logical plan to execute
            model: Optional SQLModel or Pydantic model class for result conversion

        Returns:
            QueryResult containing the execution results
        """
        stmt = self._db.compile_plan(plan)
        return self._db._executor.fetch(stmt, model=model)

    def execute_plan_stream(self, plan: "LogicalPlan") -> Iterator[List[Dict[str, object]]]:
        """Execute a plan and return an iterator of row chunks.

        Args:
            plan: Logical plan to execute

        Returns:
            Iterator of row chunks (each chunk is a list of dictionaries)
        """
        stmt = self._db.compile_plan(plan)
        return self._db._executor.fetch_stream(stmt)

    def execute_sql(self, sql: str, params: Optional[Dict[str, Any]] = None) -> "QueryResult":
        """Execute a raw SQL query.

        Args:
            sql: SQL query string
            params: Optional query parameters

        Returns:
            QueryResult containing the execution results
        """
        return self._db._executor.fetch(sql, params=params)

    def explain(self, sql: str, params: Optional[Dict[str, Any]] = None) -> str:
        """Get the execution plan for a SQL query.

        Args:
            sql: SQL query string
            params: Optional query parameters

        Returns:
            Execution plan as a string (dialect-specific)
        """
        dialect_name = self._db._dialect.name
        if dialect_name == "postgresql" or dialect_name == "duckdb":
            explain_sql = f"EXPLAIN ANALYZE {sql}"
        elif dialect_name == "mysql":
            explain_sql = f"EXPLAIN {sql}"
        elif dialect_name == "sqlite":
            explain_sql = f"EXPLAIN QUERY PLAN {sql}"
        else:
            explain_sql = f"EXPLAIN {sql}"

        result = self._db._executor.fetch(explain_sql, params=params)
        if result.rows is None:
            return ""
        # Handle different row types
        rows: List[Any] = []
        if isinstance(result.rows, list):
            rows = result.rows
        elif hasattr(result.rows, "to_dict"):
            # pandas DataFrame
            rows = result.rows.to_dict("records")  # type: ignore[call-overload]
        elif hasattr(result.rows, "to_dicts"):
            # polars DataFrame
            rows = list(result.rows.to_dicts())  # type: ignore[operator]
        # Format rows - handle dict or other types
        if not rows:
            return ""
        # Format as string
        lines = []
        for row in rows:
            if isinstance(row, dict):
                # Format dict row
                formatted = " | ".join(f"{k}: {v}" for k, v in row.items())
                lines.append(formatted)
            else:
                lines.append(str(row))
        return "\n".join(lines)



"""Compilation of math expression operations.

This module handles compilation of basic math operations like add, sub, mul, div, and math functions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from sqlalchemy import func, case as sa_case, literal

if TYPE_CHECKING:
    from sqlalchemy.sql import ColumnElement
    from ..expression_compiler import ExpressionCompiler


def compile_math_operation(
    compiler: "ExpressionCompiler", op: str, expression: Any
) -> "ColumnElement[Any] | None":
    """Compile a math operation expression.

    Args:
        compiler: The ExpressionCompiler instance
        op: Operation name (e.g., "add", "sub", "mul", "div", "abs", "sqrt", etc.)
        expression: :class:`Column` expression to compile

    Returns:
        Compiled SQLAlchemy column element, or None if not handled
    """
    result: "ColumnElement[Any]"

    # Basic arithmetic operations
    if op == "add":
        left, right = expression.args
        result = compiler._compile(left) + compiler._compile(right)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "sub":
        left, right = expression.args
        result = compiler._compile(left) - compiler._compile(right)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "mul":
        left, right = expression.args
        result = compiler._compile(left) * compiler._compile(right)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "div":
        left, right = expression.args
        result = compiler._compile(left) / compiler._compile(right)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "floor_div":
        left, right = expression.args
        result = func.floor(compiler._compile(left) / compiler._compile(right))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "round":
        col_expr = compiler._compile(expression.args[0])
        scale = expression.args[1] if len(expression.args) > 1 else 0
        result = func.round(col_expr, scale)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "floor":
        result = func.floor(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "ceil":
        col_expr = compiler._compile(expression.args[0])
        # SQLite doesn't have ceil() function, use workaround
        if compiler.dialect.name == "sqlite":
            from sqlalchemy import cast, types as sa_types

            # SQLite ceil workaround:
            # CASE WHEN x > CAST(x AS INTEGER) THEN CAST(x AS INTEGER) + 1 ELSE CAST(x AS INTEGER) END
            int_part = cast(col_expr, sa_types.Integer)
            result = sa_case(
                (col_expr > int_part, int_part + literal(1)),
                else_=int_part,
            )
        else:
            result = func.ceil(col_expr)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "abs":
        result = func.abs(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "sqrt":
        col_expr = compiler._compile(expression.args[0])
        # SQLite doesn't have sqrt() function natively
        # Some SQLite builds may have it via extensions
        # If not available, execution will fail and test should handle it
        result = func.sqrt(col_expr)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "exp":
        result = func.exp(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "log":
        col_expr = compiler._compile(expression.args[0])
        # SQLite doesn't have ln() or log() function natively
        # Some SQLite builds may have these via extensions
        if compiler.dialect.name == "sqlite":
            # Try func.ln first (SQLAlchemy may handle it if SQLite has extension)
            # If that doesn't work, the test should catch the exception
            result = func.ln(col_expr)
        else:
            result = func.ln(col_expr)  # Natural log
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "log10":
        col_expr = compiler._compile(expression.args[0])
        # SQLite doesn't have log() function with base parameter natively
        # Some SQLite builds may have log10 via extensions
        # If not available, execution will fail and test should handle it
        result = func.log(10, col_expr)  # Base-10 log
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "log2":
        col_expr = compiler._compile(expression.args[0])
        # Use dialect-specific log2
        if compiler.dialect.name == "postgresql":
            result = func.log(2, col_expr)
        elif compiler.dialect.name == "mysql":
            result = func.log(2, col_expr)
        else:
            # SQLite: log(2, x)
            result = func.log(2, col_expr)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "sin":
        result = func.sin(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "cos":
        result = func.cos(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "tan":
        result = func.tan(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "asin":
        result = func.asin(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "acos":
        result = func.acos(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "atan":
        result = func.atan(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "atan2":
        y, x = expression.args
        result = func.atan2(compiler._compile(y), compiler._compile(x))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "signum" or op == "sign":
        col_expr = compiler._compile(expression.args[0])
        # Use dialect-specific SIGN function
        if compiler.dialect.name in ("postgresql", "mysql", "duckdb"):
            result = func.sign(col_expr)
        else:
            # SQLite doesn't have SIGN, use CASE WHEN
            result = sa_case(
                (col_expr > 0, literal(1)),
                (col_expr < 0, literal(-1)),
                else_=literal(0),
            )
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "hypot":
        x, y = expression.args
        x_expr = compiler._compile(x)
        y_expr = compiler._compile(y)
        # Use dialect-specific hypot
        if compiler.dialect.name == "postgresql":
            result = func.hypot(x_expr, y_expr)
        else:
            # MySQL/SQLite: manual calculation sqrt(x² + y²)
            result = func.sqrt(x_expr * x_expr + y_expr * y_expr)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    return None  # Not handled by this module



"""Compilation of JSON expression operations.

This module handles compilation of JSON functions like json_extract, json_tuple, etc.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from sqlalchemy import func, literal

if TYPE_CHECKING:
    from sqlalchemy.sql import ColumnElement
    from ..expression_compiler import ExpressionCompiler


def compile_json_operation(
    compiler: "ExpressionCompiler", op: str, expression: Any
) -> "ColumnElement[Any] | None":
    """Compile a JSON operation expression.

    Args:
        compiler: The ExpressionCompiler instance
        op: Operation name (e.g., "json_extract", "json_tuple")
        expression: :class:`Column` expression to compile

    Returns:
        Compiled SQLAlchemy column element, or None if not handled
    """
    result: "ColumnElement[Any]"

    if op == "json_extract":
        col_expr = compiler._compile(expression.args[0])
        path_arg = expression.args[1]
        if not isinstance(path_arg, str):
            raise TypeError(f"Expected string for JSON path, got {type(path_arg).__name__}")
        path: str = path_arg
        # Use dialect-specific JSON extraction
        # PostgreSQL: -> operator or json_extract_path_text
        # SQLite: json_extract (JSON1 extension)
        # MySQL: JSON_EXTRACT or -> operator
        # Generic: Use func.json_extract which SQLAlchemy may handle
        if compiler.dialect.name == "postgresql":
            # PostgreSQL uses -> or ->> operators for JSONB
            # Convert $.key to 'key' and use -> operator
            # For paths like $.key.nested, convert to ['key', 'nested']
            if path.startswith("$."):
                # Remove $. prefix and split by . for nested paths
                path_parts = path[2:].split(".")
                # Use -> operator for JSONB (returns JSONB) or ->> for text
                # For now, use ->> to get text result
                result = col_expr
                for part in path_parts:
                    result = result.op("->>")(literal(part))
            else:
                # Use json_extract_path_text with path elements
                path_parts = path.split(".") if "." in path else [path]
                result = func.json_extract_path_text(col_expr, *[literal(p) for p in path_parts])
        elif compiler.dialect.name == "sqlite":
            # SQLite JSON1 extension
            result = func.json_extract(col_expr, path)
        else:
            # Generic fallback - try JSON_EXTRACT
            result = func.json_extract(col_expr, path)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "json_tuple":
        col_expr = compiler._compile(expression.args[0])
        # json_tuple extracts multiple keys from JSON
        # args[1:] are the keys to extract
        keys = expression.args[1:]
        if not keys:
            from ...utils.exceptions import CompilationError

            raise CompilationError("json_tuple requires at least one key argument")

        # For PostgreSQL, use json_extract_path_text for each key
        if compiler.dialect.name == "postgresql":
            from sqlalchemy import literal_column

            # Build multiple json_extract_path_text calls
            # This is a simplified implementation - full json_tuple would return multiple columns
            # For now, return the first key's value
            first_key = keys[0]
            if isinstance(first_key, str):
                result = func.json_extract_path_text(col_expr, literal(first_key))
            else:
                result = func.json_extract_path_text(col_expr, compiler._compile(first_key))
        elif compiler.dialect.name == "sqlite":
            # SQLite: json_extract for each key
            first_key = keys[0]
            if isinstance(first_key, str):
                result = func.json_extract(col_expr, f"$.{first_key}")
            else:
                result = func.json_extract(col_expr, compiler._compile(first_key))
        else:
            # Generic fallback
            first_key = keys[0]
            if isinstance(first_key, str):
                result = func.json_extract(col_expr, f"$.{first_key}")
            else:
                result = func.json_extract(col_expr, compiler._compile(first_key))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "json_array_length":
        col_expr = compiler._compile(expression.args[0])
        # Use dialect-specific array length
        if compiler.dialect.name == "postgresql":
            result = func.jsonb_array_length(col_expr)
        elif compiler.dialect.name == "mysql":
            result = func.json_length(col_expr)
        else:
            result = func.json_array_length(col_expr)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "json_tuple":
        col_expr = compiler._compile(expression.args[0])
        paths = expression.args[1:]
        if compiler.dialect.name == "postgresql":
            from sqlalchemy import literal_column

            path_list = ", ".join(f"'{p}'" for p in paths)
            result = literal_column(
                f"ARRAY(SELECT jsonb_path_query({col_expr}, p) FROM unnest(ARRAY[{path_list}]) AS p)"
            )
        else:
            results = [func.json_extract(col_expr, path) for path in paths]
            result = func.json_array(*results)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "from_json":
        col_expr = compiler._compile(expression.args[0])
        from sqlalchemy import types as sa_types

        if len(expression.args) > 1:
            # schema = expression.args[1]  # Not used in current implementation
            result = func.cast(col_expr, sa_types.JSON())
        else:
            result = func.cast(col_expr, sa_types.JSON())
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "to_json":
        col_expr = compiler._compile(expression.args[0])
        if compiler.dialect.name == "postgresql":
            result = func.to_jsonb(col_expr)
        elif compiler.dialect.name == "mysql":
            result = func.json_quote(col_expr)
        else:
            result = func.json(col_expr)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    return None  # Not handled by this module



"""Async text file reader implementation."""

from __future__ import annotations

from pathlib import Path
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    cast,
)

from ....io.records import AsyncRecords
from ....table.schema import ColumnDef
from .compression import read_compressed_async

if TYPE_CHECKING:
    from ....table.async_table import AsyncDatabase


async def read_text(
    path: str,
    database: "AsyncDatabase",
    schema: Optional[Sequence[ColumnDef]],
    options: Dict[str, object],
    column_name: str = "value",
) -> AsyncRecords:
    """Read text file line-by-line asynchronously and return :class:`AsyncRecords`.

    Args:
        path: Path to text file
        database: :class:`AsyncDatabase` instance
        schema: Optional explicit schema (unused, always TEXT)
        options: Reader options (unused for text)
        column_name: Name of the column to create (default: "value")

    Returns:
        :class:`AsyncRecords` containing the text file lines

    Raises:
        FileNotFoundError: If file doesn't exist
    """
    path_obj = Path(path)
    if not path_obj.exists():
        raise FileNotFoundError(f"Text file not found: {path}")

    compression = cast(Optional[str], options.get("compression", None))
    encoding = cast(str, options.get("encoding", "UTF-8"))
    wholetext = cast(bool, options.get("wholetext", False))
    line_sep = cast(Optional[str], options.get("lineSep", None))

    content = await read_compressed_async(path, compression=compression, encoding=encoding)
    rows: List[Dict[str, object]] = []

    if wholetext:
        # Read entire file as single value
        rows.append({column_name: content})
    else:
        # Read line by line
        if line_sep:
            lines = content.split(line_sep)
            for line in lines:
                rows.append({column_name: line})
        else:
            for line in content.splitlines(keepends=True):
                rows.append({column_name: line.rstrip("\n\r")})

    if not rows:
        return _create_async_records_from_schema(database, [], [])

    schema = [ColumnDef(name=column_name, type_name="TEXT", nullable=False)]
    return _create_async_records_from_data(database, rows, schema)


async def read_text_stream(
    path: str,
    database: "AsyncDatabase",
    schema: Optional[Sequence[ColumnDef]],
    options: Dict[str, object],
    column_name: str = "value",
) -> AsyncRecords:
    """Read text file asynchronously in streaming mode (chunked).

    Args:
        path: Path to text file
        database: :class:`AsyncDatabase` instance
        schema: Optional explicit schema (unused, always TEXT)
        options: Reader options (chunk_size)
        column_name: Name of the column to create (default: "value")

    Returns:
        :class:`AsyncRecords` with streaming generator

    Raises:
        FileNotFoundError: If file doesn't exist
    """
    chunk_size = int(cast(int, options.get("chunk_size", 10000)))
    path_obj = Path(path)
    if not path_obj.exists():
        raise FileNotFoundError(f"Text file not found: {path}")

    compression = cast(Optional[str], options.get("compression", None))
    encoding = cast(str, options.get("encoding", "UTF-8"))
    wholetext = cast(bool, options.get("wholetext", False))
    line_sep = cast(Optional[str], options.get("lineSep", None))

    async def _chunk_generator() -> AsyncIterator[List[Dict[str, object]]]:
        content = await read_compressed_async(path, compression=compression, encoding=encoding)
        chunk: List[Dict[str, object]] = []

        if wholetext:
            # Read entire file as single value
            yield [{column_name: content}]
        else:
            # Read line by line
            if line_sep:
                lines = content.split(line_sep)
                for line in lines:
                    chunk.append({column_name: line})
                    if len(chunk) >= chunk_size:
                        yield chunk
                        chunk = []
                if chunk:
                    yield chunk
            else:
                for line in content.splitlines(keepends=True):
                    chunk.append({column_name: line.rstrip("\n\r")})
                    if len(chunk) >= chunk_size:
                        yield chunk
                        chunk = []
                if chunk:
                    yield chunk

    # Read first chunk
    first_chunk_gen = _chunk_generator()
    try:
        first_chunk = await first_chunk_gen.__anext__()
    except StopAsyncIteration:
        schema = [ColumnDef(name=column_name, type_name="TEXT", nullable=False)]
        return _create_async_records_from_schema(database, schema, [])

    schema = [ColumnDef(name=column_name, type_name="TEXT", nullable=False)]

    async def _typed_chunk_generator() -> AsyncIterator[List[Dict[str, object]]]:
        yield first_chunk
        async for chunk in first_chunk_gen:
            yield chunk

    return _create_async_records_from_stream(database, _typed_chunk_generator, schema)


def _create_async_records_from_data(
    database: "AsyncDatabase", rows: List[Dict[str, object]], schema: Optional[Sequence[ColumnDef]]
) -> AsyncRecords:
    """Create :class:`AsyncRecords` from materialized data."""
    return AsyncRecords(_data=rows, _schema=schema, _database=database)


def _create_async_records_from_schema(
    database: "AsyncDatabase", schema: Sequence[ColumnDef], rows: List[Dict[str, object]]
) -> AsyncRecords:
    """Create :class:`AsyncRecords` with explicit schema but no data."""
    return AsyncRecords(_data=rows, _schema=schema, _database=database)


def _create_async_records_from_stream(
    database: "AsyncDatabase",
    chunk_generator: Callable[[], AsyncIterator[List[Dict[str, object]]]],
    schema: Sequence[ColumnDef],
) -> AsyncRecords:
    """Create :class:`AsyncRecords` from streaming generator."""
    return AsyncRecords(_generator=chunk_generator, _schema=schema, _database=database)



"""Compilation of logical expression operations.

This module handles compilation of logical operations like and, or, not, between, in, case_when, etc.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from sqlalchemy import and_, or_, not_, case as sa_case

if TYPE_CHECKING:
    from sqlalchemy.sql import ColumnElement
    from ..expression_compiler import ExpressionCompiler


def compile_logical_operation(
    compiler: "ExpressionCompiler", op: str, expression: Any
) -> "ColumnElement[Any] | None":
    """Compile a logical operation expression.

    Args:
        compiler: The ExpressionCompiler instance
        op: Operation name (e.g., "and", "or", "not", "between", "in", "case_when")
        expression: :class:`Column` expression to compile

    Returns:
        Compiled SQLAlchemy column element, or None if not handled
    """
    result: "ColumnElement[Any]"

    if op == "and":
        left, right = expression.args
        result = and_(compiler._compile(left), compiler._compile(right))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "or":
        left, right = expression.args
        result = or_(compiler._compile(left), compiler._compile(right))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "not":
        result = not_(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "between":
        value, lower, upper = expression.args
        result = compiler._compile(value).between(
            compiler._compile(lower), compiler._compile(upper)
        )
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "in":
        value, options = expression.args
        if not isinstance(options, (list, tuple)):
            raise TypeError(f"Expected iterable for 'in' options, got {type(options).__name__}")
        option_values = [compiler._compile(opt) for opt in options]
        result = compiler._compile(value).in_(option_values)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "case_when":
        # CASE WHEN expression: args[0] is tuple of (condition, value) pairs, args[1] is else value
        conditions = expression.args[0]
        if not isinstance(conditions, (list, tuple)):
            raise TypeError(
                f"Expected iterable for CASE conditions, got {type(conditions).__name__}"
            )
        else_value = compiler._compile(expression.args[1])

        # Build CASE statement
        # Start with empty when clauses, add them one by one
        when_clauses: list[tuple[ColumnElement[Any], Any]] = []
        for condition, value in conditions:
            when_clauses.append((compiler._compile(condition), compiler._compile(value)))
        case_stmt = sa_case(*when_clauses, else_=else_value)

        result = case_stmt
        if expression._alias:
            result = result.label(expression._alias)
        return result

    return None  # Not handled by this module



"""Django integration utilities for Moltres.

This module provides helper functions and utilities to make Moltres
more user-friendly and robust when used with Django.

Key features:
- Middleware for converting Moltres errors to Django HTTP responses
- :class:`Database` connection helpers with Django database routing support
- Integration with Django's transaction management
- Logging integration with Django's logging system
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Callable, Optional, cast

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse, JsonResponse  # type: ignore[import-untyped]
    from django.utils.deprecation import MiddlewareMixin  # type: ignore[import-untyped]
    from ...table.table import Database
else:
    Database = Any
    HttpRequest = Any
    HttpResponse = Any
    JsonResponse = Any

try:
    from django.conf import settings  # type: ignore[import-untyped]
    from django.core.exceptions import ImproperlyConfigured  # type: ignore[import-untyped]
    from django.db import connections, transaction  # type: ignore[import-untyped]
    from django.http import HttpRequest, HttpResponse, JsonResponse
    from django.utils.deprecation import MiddlewareMixin

    DJANGO_AVAILABLE = True
except ImportError:
    DJANGO_AVAILABLE = False
    # Create stubs for type checking
    MiddlewareMixin = cast(type[Any], None)
    HttpRequest = cast(type[Any], None)
    HttpResponse = cast(type[Any], None)
    JsonResponse = cast(type[Any], None)
    connections = None
    transaction = None
    settings = None
    ImproperlyConfigured = None


logger = logging.getLogger(__name__)


class MoltresExceptionMiddleware:
    """Django middleware that catches Moltres exceptions and converts them to HTTP responses.

    This middleware should be added to Django's MIDDLEWARE setting to automatically
    handle Moltres exceptions throughout your application.

    Example:
        Add to settings.py:
        MIDDLEWARE = [
            # ... other middleware
            'moltres.integrations.django.MoltresExceptionMiddleware',
        ]
    """

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]) -> None:
        """Initialize the middleware.

        Args:
            get_response: The next middleware or view in the chain
        """
        if not DJANGO_AVAILABLE:
            raise ImportError(
                "Django is required for MoltresExceptionMiddleware. Install with: pip install django"
            )
        self.get_response = get_response
        self._setup_logging()

    def _setup_logging(self) -> None:
        """Set up logging for Moltres exceptions."""
        # Use Django's logging configuration
        if not hasattr(logger, "handlers") or not logger.handlers:
            # Add a handler if none exists
            handler = logging.StreamHandler()
            formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
            handler.setFormatter(formatter)
            logger.addHandler(handler)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        """Process the request and handle Moltres exceptions.

        Args:
            request: Django HttpRequest object

        Returns:
            HttpResponse with appropriate status code and error details
        """
        try:
            response = self.get_response(request)
            return response
        except Exception as exc:
            return self._handle_exception(request, exc)

    def _handle_exception(self, request: HttpRequest, exc: Exception) -> HttpResponse:
        """Handle Moltres exceptions and convert to HTTP responses.

        Args:
            request: Django HttpRequest object
            exc: Exception that was raised

        Returns:
            JsonResponse with appropriate status code and error details
        """
        from ...utils.exceptions import (
            ColumnNotFoundError,
            CompilationError,
            ConnectionPoolError,
            DatabaseConnectionError,
            ExecutionError,
            MoltresError,
            QueryTimeoutError,
            TransactionError,
            ValidationError,
        )

        # Log the exception
        logger.error(
            "Moltres exception in Django view",
            exc_info=True,
            extra={"request_path": request.path, "exception_type": type(exc).__name__},
        )

        # Handle specific Moltres exceptions
        # Check more specific exceptions first (ConnectionPoolError is a subclass of DatabaseConnectionError)
        if isinstance(exc, ConnectionPoolError):
            return JsonResponse(
                {
                    "error": "Connection pool error",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "detail": exc.context,
                },
                status=503,  # Service Unavailable
            )

        if isinstance(exc, DatabaseConnectionError):
            return JsonResponse(
                {
                    "error": "Database connection error",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "detail": exc.context,
                },
                status=503,  # Service Unavailable
            )

        if isinstance(exc, QueryTimeoutError):
            return JsonResponse(
                {
                    "error": "Query timeout",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "timeout_seconds": exc.context.get("timeout_seconds") if exc.context else None,
                    "detail": exc.context,
                },
                status=504,  # Gateway Timeout
            )

        # Check more specific ExecutionError subclasses first
        if isinstance(exc, QueryTimeoutError):
            return JsonResponse(
                {
                    "error": "Query timeout",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "timeout_seconds": exc.context.get("timeout_seconds") if exc.context else None,
                    "detail": exc.context,
                },
                status=504,  # Gateway Timeout
            )

        if isinstance(exc, TransactionError):
            return JsonResponse(
                {
                    "error": "Transaction error",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "detail": exc.context,
                },
                status=500,  # Internal Server Error
            )

        if isinstance(exc, ExecutionError):
            # Check if it's a "not found" type error
            error_msg = str(exc.message).lower()
            if "not found" in error_msg or "does not exist" in error_msg:
                status_code = 404  # Not Found
            elif "permission" in error_msg or "access" in error_msg:
                status_code = 403  # Forbidden
            elif "syntax error" in error_msg or "invalid" in error_msg:
                status_code = 400  # Bad Request
            else:
                status_code = 500  # Internal Server Error

            return JsonResponse(
                {
                    "error": "SQL execution error",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "detail": exc.context,
                },
                status=status_code,
            )

        if isinstance(exc, CompilationError):
            return JsonResponse(
                {
                    "error": "SQL compilation error",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "detail": exc.context,
                },
                status=400,  # Bad Request
            )

        # Check more specific ValidationError subclasses first
        if isinstance(exc, ColumnNotFoundError):
            return JsonResponse(
                {
                    "error": "Column not found",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "column_name": exc.context.get("column_name") if exc.context else None,
                    "available_columns": exc.context.get("available_columns")
                    if exc.context
                    else None,
                },
                status=400,  # Bad Request
            )

        if isinstance(exc, ValidationError):
            return JsonResponse(
                {
                    "error": "Validation error",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "detail": exc.context,
                },
                status=400,  # Bad Request
            )

        if isinstance(exc, MoltresError):
            return JsonResponse(
                {
                    "error": "Moltres error",
                    "message": str(exc.message),
                    "suggestion": exc.suggestion,
                    "detail": exc.context,
                },
                status=500,  # Internal Server Error
            )

        # Not a Moltres exception, re-raise it
        raise exc


def get_moltres_db(using: str = "default") -> Database:
    """Get a Moltres :class:`Database` instance from Django's database connection.

    This function creates a Moltres :class:`Database` instance using Django's database
    connection, supporting Django's database routing and transaction management.

    Args:
        using: Django database alias (default: 'default'). Supports database routing.

    Returns:
        Moltres :class:`Database` instance

    Raises:
        ImportError: If Django is not installed
        ImproperlyConfigured: If the database alias is not configured

    Example:
        >>> from moltres.integrations.django import get_moltres_db
        >>>
        >>> def my_view(request):
        ...     db = get_moltres_db(using='default')
        ...     df = db.table("users").select()
        ...     return JsonResponse({'users': df.collect()})
    """
    if not DJANGO_AVAILABLE:
        raise ImportError("Django is required for get_moltres_db. Install with: pip install django")

    # Validate database alias
    if using not in settings.DATABASES:
        raise ImproperlyConfigured(
            f"Database alias '{using}' is not configured in Django settings.DATABASES"
        )

    # Get connection parameters from Django settings
    db_config = settings.DATABASES[using]
    engine_name = db_config.get("ENGINE", "").lower()

    # Build SQLAlchemy connection string from Django settings
    dsn: Optional[str] = None

    if "sqlite" in engine_name:
        # SQLite
        name = db_config.get("NAME", ":memory:")
        # Handle both absolute and relative paths
        if name == ":memory:":
            dsn = "sqlite:///:memory:"
        else:
            dsn = f"sqlite:///{name}"
    elif "postgresql" in engine_name or "postgis" in engine_name:
        # PostgreSQL
        user = db_config.get("USER", "")
        password = db_config.get("PASSWORD", "")
        host = db_config.get("HOST", "localhost")
        port = db_config.get("PORT", "5432")
        name = db_config.get("NAME", "")
        # URL encode password if it contains special characters
        from urllib.parse import quote_plus

        password_encoded = quote_plus(password) if password else ""
        dsn = f"postgresql://{user}:{password_encoded}@{host}:{port}/{name}"
    elif "mysql" in engine_name:
        # MySQL
        user = db_config.get("USER", "")
        password = db_config.get("PASSWORD", "")
        host = db_config.get("HOST", "localhost")
        port = db_config.get("PORT", "3306")
        name = db_config.get("NAME", "")
        # URL encode password if it contains special characters
        from urllib.parse import quote_plus

        password_encoded = quote_plus(password) if password else ""
        dsn = f"mysql://{user}:{password_encoded}@{host}:{port}/{name}"
    else:
        raise ImproperlyConfigured(
            f"Unsupported database engine '{engine_name}' for database '{using}'. "
            "Moltres Django integration currently supports: sqlite3, postgresql, mysql. "
            "For other databases, use connect() directly with a DSN string."
        )

    # Create Moltres Database instance
    from ... import connect

    db = connect(dsn=dsn)

    # Store reference to Django connection for transaction management
    # This allows us to integrate with Django's transaction.atomic()
    django_connection = connections[using]
    db._django_connection = django_connection  # type: ignore[attr-defined]
    db._django_using = using  # type: ignore[attr-defined]

    return db



"""I/O operations for DataFrame (readers and writers)."""

from .reader import DataLoader, ReadAccessor
from .async_reader import AsyncDataLoader, AsyncReadAccessor
from .writer import DataFrameWriter
from .async_writer import AsyncDataFrameWriter

__all__ = [
    "DataLoader",
    "ReadAccessor",
    "AsyncDataLoader",
    "AsyncReadAccessor",
    "DataFrameWriter",
    "AsyncDataFrameWriter",
]



"""Transaction metrics collection and monitoring.

This module provides metrics collection for transaction operations, tracking
statistics such as transaction count, duration, success/failure rates, etc.
"""

from __future__ import annotations

import time
from dataclasses import dataclass, field
from typing import Any, Dict, Optional


@dataclass
class TransactionMetrics:
    """Metrics for transaction operations."""

    transaction_count: int = 0
    transaction_duration_sum: float = 0.0
    transaction_duration_min: float = float("inf")
    transaction_duration_max: float = 0.0
    committed_count: int = 0
    rolled_back_count: int = 0
    savepoint_count: int = 0
    readonly_count: int = 0
    isolation_levels: Dict[str, int] = field(default_factory=dict)
    error_count: int = 0
    errors_by_type: Dict[str, int] = field(default_factory=dict)

    def record_transaction(
        self,
        duration: float,
        committed: bool = True,
        has_savepoint: bool = False,
        readonly: bool = False,
        isolation_level: Optional[str] = None,
        error: Optional[Exception] = None,
    ) -> None:
        """Record a transaction execution.

        Args:
            duration: Transaction duration in seconds
            committed: Whether transaction was committed (True) or rolled back (False)
            has_savepoint: Whether transaction used savepoints
            readonly: Whether transaction was read-only
            isolation_level: Transaction isolation level if set
            error: Exception if transaction failed
        """
        self.transaction_count += 1
        self.transaction_duration_sum += duration
        self.transaction_duration_min = min(self.transaction_duration_min, duration)
        self.transaction_duration_max = max(self.transaction_duration_max, duration)

        if committed:
            self.committed_count += 1
        else:
            self.rolled_back_count += 1

        if has_savepoint:
            self.savepoint_count += 1

        if readonly:
            self.readonly_count += 1

        if isolation_level:
            self.isolation_levels[isolation_level] = (
                self.isolation_levels.get(isolation_level, 0) + 1
            )

        if error:
            self.error_count += 1
            error_type = type(error).__name__
            self.errors_by_type[error_type] = self.errors_by_type.get(error_type, 0) + 1

    def get_stats(self) -> Dict[str, Any]:
        """Get current transaction statistics.

        Returns:
            Dictionary with transaction statistics
        """
        avg_duration = (
            self.transaction_duration_sum / self.transaction_count
            if self.transaction_count > 0
            else 0.0
        )

        return {
            "transaction_count": self.transaction_count,
            "transaction_duration_avg": avg_duration,
            "transaction_duration_min": self.transaction_duration_min
            if self.transaction_duration_min != float("inf")
            else 0.0,
            "transaction_duration_max": self.transaction_duration_max,
            "committed_count": self.committed_count,
            "rolled_back_count": self.rolled_back_count,
            "commit_rate": (
                self.committed_count / self.transaction_count if self.transaction_count > 0 else 0.0
            ),
            "savepoint_count": self.savepoint_count,
            "readonly_count": self.readonly_count,
            "isolation_levels": dict(self.isolation_levels),
            "error_count": self.error_count,
            "error_rate": (
                self.error_count / self.transaction_count if self.transaction_count > 0 else 0.0
            ),
            "errors_by_type": dict(self.errors_by_type),
        }

    def reset(self) -> None:
        """Reset all metrics."""
        self.transaction_count = 0
        self.transaction_duration_sum = 0.0
        self.transaction_duration_min = float("inf")
        self.transaction_duration_max = 0.0
        self.committed_count = 0
        self.rolled_back_count = 0
        self.savepoint_count = 0
        self.readonly_count = 0
        self.isolation_levels.clear()
        self.error_count = 0
        self.errors_by_type.clear()


# Global transaction metrics instance
_global_transaction_metrics = TransactionMetrics()


def get_transaction_metrics() -> TransactionMetrics:
    """Get the global transaction metrics collector.

    Returns:
        TransactionMetrics instance
    """
    return _global_transaction_metrics


def reset_transaction_metrics() -> None:
    """Reset global transaction metrics."""
    _global_transaction_metrics.reset()


class TransactionContext:
    """Context manager for tracking transaction metrics."""

    def __init__(
        self,
        has_savepoint: bool = False,
        readonly: bool = False,
        isolation_level: Optional[str] = None,
    ):
        """Initialize transaction context.

        Args:
            has_savepoint: Whether transaction uses savepoints
            readonly: Whether transaction is read-only
            isolation_level: Transaction isolation level if set
        """
        self.has_savepoint = has_savepoint
        self.readonly = readonly
        self.isolation_level = isolation_level
        self.start_time: Optional[float] = None
        self.committed = False
        self.error: Optional[Exception] = None

    def __enter__(self) -> "TransactionContext":
        """Enter transaction context."""
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Exit transaction context and record metrics."""
        if self.start_time is None:
            return

        duration = time.time() - self.start_time
        # Use committed flag if set, otherwise determine from exc_type and error
        if hasattr(self, "committed") and self.committed:
            committed = True
        elif exc_type is not None or self.error is not None:
            committed = False
        else:
            committed = True  # No exception and no error means committed

        _global_transaction_metrics.record_transaction(
            duration=duration,
            committed=committed,
            has_savepoint=self.has_savepoint,
            readonly=self.readonly,
            isolation_level=self.isolation_level,
            error=exc_val if exc_type else self.error,
        )

    def mark_committed(self) -> None:
        """Mark transaction as committed."""
        self.committed = True

    def mark_error(self, error: Exception) -> None:
        """Mark transaction as having an error."""
        self.error = error



"""Json functions for :class:`DataFrame` operations."""

from __future__ import annotations

from typing import Optional

from ..column import Column, ColumnLike, ensure_column


def json_extract(column: ColumnLike, path: str) -> Column:
    """Extract a value from a JSON column using a JSON path.

    Args:
        column: JSON column expression
        path: JSON path expression (e.g., "$.key", "$.nested.key", "$[0]")

    Returns:
        :class:`Column` expression for json_extract

    Example:
        >>> # Note: json_extract() requires database-specific JSON support
        >>> # SQLite has limited JSON support via json_extract() function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("sqlite:///:memory:")
        >>> _ = db.create_table("data", [column("json_data", "TEXT")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"json_data": '{"name": "Alice", "age": 30}'}], _database=db).insert_into("data")
        >>> df = db.table("data").select(F.json_extract(col("json_data"), "$.name").alias("name"))
        >>> results = df.collect()
        >>> results[0]["name"]
        'Alice'
        >>> db.close()
    """
    return Column(op="json_extract", args=(ensure_column(column), path))


def json_tuple(column: ColumnLike, *paths: str) -> Column:
    """Extract multiple JSON paths from a JSON column at once.

    Args:
        column: JSON column expression
        *paths: JSON path expressions (e.g., "$.key1", "$.key2")

    Returns:
        :class:`Column` expression for json_tuple (returns array of values)

    Example:
        >>> # Note: json_tuple() requires database-specific JSON support (PostgreSQL/MySQL)
        >>> # SQLite does not have json_tuple function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("data", [column("json_data", "TEXT")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"json_data": '{"name": "Alice", "age": 30}'}], _database=db).insert_into("data")
        >>> df = db.table("data").select(F.json_tuple(col("json_data"), "$.name", "$.age").alias("tuple"))
        >>> results = df.collect()
        >>> len(results[0]["tuple"])
        2
        >>> db.close()
    """
    if not paths:
        raise ValueError("json_tuple requires at least one path")
    return Column(op="json_tuple", args=(ensure_column(column),) + paths)


def from_json(column: ColumnLike, schema: Optional[str] = None) -> Column:
    """Parse a JSON string column into a JSON object.

    Args:
        column: String column containing JSON
        schema: Optional schema string (for validation)

    Returns:
        :class:`Column` expression for from_json

    Example:
        >>> # Note: from_json() requires database-specific JSON support (PostgreSQL/MySQL)
        >>> # SQLite does not have from_json function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("data", [column("json_str", "TEXT")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"json_str": '{"name": "Alice"}'}], _database=db).insert_into("data")
        >>> df = db.table("data").select(F.from_json(col("json_str")).alias("json_obj"))
        >>> results = df.collect()
        >>> isinstance(results[0]["json_obj"], dict)
        True
        >>> db.close()
    """
    if schema is not None:
        return Column(op="from_json", args=(ensure_column(column), schema))
    return Column(op="from_json", args=(ensure_column(column),))


def to_json(column: ColumnLike) -> Column:
    """Convert a column to a JSON string.

    Args:
        column: :class:`Column` expression to convert

    Returns:
        :class:`Column` expression for to_json

    Example:
        >>> # Note: to_json() requires database-specific JSON support (PostgreSQL/MySQL)
        >>> # SQLite does not have to_json function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("duckdb:///:memory:")
        >>> _ = db.create_table("data", [column("value", "INTEGER")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"value": 42}], _database=db).insert_into("data")
        >>> df = db.table("data").select(F.to_json(col("value")).alias("json_str"))
        >>> results = df.collect()
        >>> # DuckDB returns JSON as the actual value type, not a string
        >>> results[0]["json_str"] in (42, '42', '"42"')
        True
        >>> db.close()
    """
    return Column(op="to_json", args=(ensure_column(column),))


def json_array_length(column: ColumnLike) -> Column:
    """Get the length of a JSON array.

    Args:
        column: JSON array column expression

    Returns:
        :class:`Column` expression for json_array_length

    Example:
        >>> # Note: json_array_length() requires database-specific JSON support
        >>> # SQLite has json_array_length() function
        >>> from moltres import connect, col
        >>> from moltres.expressions import functions as F
        >>> from moltres.table.schema import column
        >>> db = connect("sqlite:///:memory:")
        >>> _ = db.create_table("data", [column("json_array", "TEXT")]).collect()
        >>> from moltres.io.records import :class:`Records`
        >>> _ = :class:`Records`(_data=[{"json_array": '[1, 2, 3]'}], _database=db).insert_into("data")
        >>> df = db.table("data").select(F.json_array_length(col("json_array")).alias("length"))
        >>> results = df.collect()
        >>> results[0]["length"]
        3
        >>> db.close()
    """
    return Column(op="json_array_length", args=(ensure_column(column),))



"""Async data loading operations."""

from __future__ import annotations

from typing import cast, TYPE_CHECKING, Any, Dict, Optional, Sequence

from ...io.records import AsyncLazyRecords, AsyncRecords
from ...table.schema import ColumnDef
from ..core.async_dataframe import AsyncDataFrame
from .readers.async_csv_reader import read_csv, read_csv_stream
from .readers.async_json_reader import (
    read_json,
    read_json_stream,
    read_jsonl,
    read_jsonl_stream,
)
from .readers.async_text_reader import read_text, read_text_stream


# Lazy import for parquet (optional dependency)
def _get_parquet_readers() -> tuple[Any, Any]:
    """Lazy import for parquet readers."""
    try:
        from .readers.async_parquet_reader import read_parquet, read_parquet_stream

        return read_parquet, read_parquet_stream
    except ImportError:
        return None, None


if TYPE_CHECKING:
    from ...table.async_table import AsyncDatabase


class AsyncDataLoader:
    """Builder for loading data from files and tables as AsyncDataFrames."""

    def __init__(self, database: "AsyncDatabase"):
        self._database = database
        self._schema: Optional[Sequence[ColumnDef]] = None
        self._options: Dict[str, object] = {}

    def stream(self, enabled: bool = True) -> "AsyncDataLoader":
        """Enable or disable streaming mode (chunked reading for large files)."""
        from ..helpers.reader_helpers import build_stream_setter

        return cast("AsyncDataLoader", build_stream_setter(self, enabled))

    def schema(self, schema: Sequence[ColumnDef]) -> "AsyncDataLoader":
        """Set an explicit schema for the data source."""
        from ..helpers.reader_helpers import build_schema_setter

        return cast("AsyncDataLoader", build_schema_setter(self, schema))

    def option(self, key: str, value: object) -> "AsyncDataLoader":
        """Set a read option (e.g., header=True for CSV, multiline=True for JSON)."""
        from ..helpers.reader_helpers import build_option_setter

        return cast("AsyncDataLoader", build_option_setter(self, key, value))

    def options(self, **options: object) -> "AsyncDataLoader":
        """Set multiple read options at once (PySpark-compatible).

        Args:
            **options: Dictionary of option key-value pairs

        Returns:
            Self for method chaining

        Example:
            >>> df = await db.read.options(header=True, delimiter=",").csv("data.csv")
        """
        from ..helpers.reader_helpers import build_options_setter

        return cast("AsyncDataLoader", build_options_setter(self, **options))

    async def table(self, name: str) -> AsyncDataFrame:
        """Read from a database table as an AsyncDataFrame.

        Note: This is equivalent to await db.table(name).select().
        Returns an AsyncDataFrame that can be transformed before execution.
        """
        table_handle = await self._database.table(name)
        return table_handle.select()

    async def csv(self, path: str) -> AsyncDataFrame:
        """Read a CSV file as an AsyncDataFrame.

        Args:
            path: Path to the CSV file

        Returns:
            AsyncDataFrame containing the CSV data (lazy, materialized on .collect())
        """
        from ..helpers.reader_helpers import build_file_scan_async_dataframe

        return build_file_scan_async_dataframe(self, path, "csv")

    async def json(self, path: str) -> AsyncDataFrame:
        """Read a JSON file (array of objects) as an AsyncDataFrame.

        Args:
            path: Path to the JSON file

        Returns:
            AsyncDataFrame containing the JSON data (lazy, materialized on .collect())
        """
        from ..helpers.reader_helpers import build_file_scan_async_dataframe

        return build_file_scan_async_dataframe(self, path, "json")

    async def jsonl(self, path: str) -> AsyncDataFrame:
        """Read a JSONL file (one JSON object per line) as an AsyncDataFrame.

        Args:
            path: Path to the JSONL file

        Returns:
            AsyncDataFrame containing the JSONL data (lazy, materialized on .collect())
        """
        from ..helpers.reader_helpers import build_file_scan_async_dataframe

        return build_file_scan_async_dataframe(self, path, "jsonl")

    async def parquet(self, path: str) -> AsyncDataFrame:
        """Read a Parquet file as an AsyncDataFrame.

        Args:
            path: Path to the Parquet file

        Returns:
            AsyncDataFrame containing the Parquet data (lazy, materialized on .collect())

        Raises:
            RuntimeError: If pandas or pyarrow are not installed
        """
        from ..helpers.reader_helpers import build_file_scan_async_dataframe

        return build_file_scan_async_dataframe(self, path, "parquet")

    async def text(self, path: str, column_name: str = "value") -> AsyncDataFrame:
        """Read a text file as a single column (one line per row) as an AsyncDataFrame.

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncDataFrame containing the text file lines (lazy, materialized on .collect())
        """
        from ..helpers.reader_helpers import build_file_scan_async_dataframe

        return build_file_scan_async_dataframe(self, path, "text", column_name=column_name)

    async def textFile(self, path: str, column_name: str = "value") -> AsyncDataFrame:
        """Read a text file as a single column (PySpark-compatible alias for text()).

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncDataFrame containing the text file lines (lazy, materialized on .collect())
        """
        return await self.text(path, column_name)

    async def format(self, source: str) -> "AsyncFormatReader":
        """Specify the data source format.

        Args:
            source: Format name (e.g., "csv", "json", "parquet")

        Returns:
            AsyncFormatReader for the specified format
        """
        from ..helpers.reader_helpers import validate_format

        return AsyncFormatReader(self, validate_format(source))


class AsyncFormatReader:
    """Builder for format-specific async reads."""

    def __init__(self, reader: AsyncDataLoader, source: str):
        self._reader = reader
        self._source = source

    async def load(self, path: str) -> AsyncDataFrame:
        """Load data from the specified path using the configured format.

        Args:
            path: Path to the data file

        Returns:
            AsyncDataFrame containing the data (lazy, materialized on .collect())

        Raises:
            ValueError: If format is unsupported
        """
        from ..helpers.reader_helpers import validate_format

        format_name = validate_format(self._source)
        if format_name == "csv":
            return await self._reader.csv(path)
        elif format_name == "json":
            return await self._reader.json(path)
        elif format_name == "jsonl":
            return await self._reader.jsonl(path)
        elif format_name == "parquet":
            read_parquet, _ = _get_parquet_readers()
            if read_parquet is None:
                raise ImportError(
                    "Parquet support requires pyarrow. Install with: pip install pyarrow"
                )
            return await self._reader.parquet(path)
        elif format_name == "text":
            return await self._reader.text(path)
        else:
            raise ValueError(f"Unsupported format: {format_name}")


class AsyncRecordsLoader:
    """Builder for loading data from files as AsyncLazyRecords (lazy :class:`AsyncRecords`).

    Provides backward compatibility and convenience for cases where :class:`AsyncRecords` are preferred
    over AsyncDataFrames. Use await db.read.records.csv() etc. to get AsyncLazyRecords directly.
    AsyncLazyRecords materialize on-demand when used.
    """

    def __init__(self, database: "AsyncDatabase"):
        self._database = database
        self._schema: Optional[Sequence[ColumnDef]] = None
        self._options: Dict[str, object] = {}

    def stream(self, enabled: bool = True) -> "AsyncRecordsLoader":
        """Enable or disable streaming mode (chunked reading for large files)."""
        from ..helpers.reader_helpers import build_stream_setter

        return cast("AsyncRecordsLoader", build_stream_setter(self, enabled))

    def schema(self, schema: Sequence[ColumnDef]) -> "AsyncRecordsLoader":
        """Set an explicit schema for the data source."""
        from ..helpers.reader_helpers import build_schema_setter

        return cast("AsyncRecordsLoader", build_schema_setter(self, schema))

    def option(self, key: str, value: object) -> "AsyncRecordsLoader":
        """Set a read option (e.g., header=True for CSV, multiline=True for JSON)."""
        from ..helpers.reader_helpers import build_option_setter

        return cast("AsyncRecordsLoader", build_option_setter(self, key, value))

    def options(self, **options: object) -> "AsyncRecordsLoader":
        """Set multiple read options at once (PySpark-compatible).

        Args:
            **options: Dictionary of option key-value pairs

        Returns:
            Self for method chaining
        """
        from ..helpers.reader_helpers import build_options_setter

        return cast("AsyncRecordsLoader", build_options_setter(self, **options))

    def csv(self, path: str) -> AsyncLazyRecords:
        """Read a CSV file as AsyncLazyRecords.

        Args:
            path: Path to the CSV file

        Returns:
            AsyncLazyRecords containing the CSV data (materializes on-demand)
        """

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return await read_csv_stream(path, self._database, self._schema, self._options)
            return await read_csv(path, self._database, self._schema, self._options)

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def json(self, path: str) -> AsyncLazyRecords:
        """Read a JSON file (array of objects) as AsyncLazyRecords.

        Args:
            path: Path to the JSON file

        Returns:
            AsyncLazyRecords containing the JSON data (materializes on-demand)
        """

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return await read_json_stream(path, self._database, self._schema, self._options)
            return await read_json(path, self._database, self._schema, self._options)

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def jsonl(self, path: str) -> AsyncLazyRecords:
        """Read a JSONL file (one JSON object per line) as AsyncLazyRecords.

        Args:
            path: Path to the JSONL file

        Returns:
            AsyncLazyRecords containing the JSONL data (materializes on-demand)
        """

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return await read_jsonl_stream(path, self._database, self._schema, self._options)
            return await read_jsonl(path, self._database, self._schema, self._options)

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def parquet(self, path: str) -> AsyncLazyRecords:
        """Read a Parquet file as AsyncLazyRecords.

        Args:
            path: Path to the Parquet file

        Returns:
            AsyncLazyRecords containing the Parquet data (materializes on-demand)

        Raises:
            RuntimeError: If pandas or pyarrow are not installed
        """
        read_parquet, read_parquet_stream = _get_parquet_readers()
        if read_parquet is None:
            raise ImportError("Parquet support requires pyarrow. Install with: pip install pyarrow")

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return cast(
                    "AsyncRecords",
                    await read_parquet_stream(path, self._database, self._schema, self._options),
                )
            return cast(
                "AsyncRecords",
                await read_parquet(path, self._database, self._schema, self._options),
            )

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def text(self, path: str, column_name: str = "value") -> AsyncLazyRecords:
        """Read a text file as a single column (one line per row) as AsyncLazyRecords.

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncLazyRecords containing the text file lines (materializes on-demand)
        """

        async def read_func() -> AsyncRecords:
            stream = self._options.get("stream", False)
            if stream:
                return await read_text_stream(
                    path, self._database, self._schema, self._options, column_name
                )
            return await read_text(path, self._database, self._schema, self._options, column_name)

        return AsyncLazyRecords(
            _read_func=read_func,
            _database=self._database,
            _schema=self._schema,
            _options=self._options.copy(),
        )

    def dicts(self, data: Sequence[Dict[str, object]]) -> AsyncRecords:
        """Create :class:`AsyncRecords` from a list of dictionaries.

        Note: This returns :class:`AsyncRecords` (not AsyncLazyRecords) since the data is already materialized.

        Args:
            data: List of dictionaries to convert to :class:`AsyncRecords`

        Returns:
            :class:`AsyncRecords` containing the data (already materialized)
        """
        return AsyncRecords(_data=list(data), _database=self._database, _schema=self._schema)


class AsyncReadAccessor:
    """Accessor for async read operations.

    Provides PySpark-style API: await db.read.table(), await db.read.csv(), etc.
    Also provides backward compatibility via db.read.records.*
    """

    def __init__(self, database: "AsyncDatabase"):
        self._database = database
        self._loader = AsyncDataLoader(database)
        self._records = AsyncRecordsLoader(database)

    @property
    def records(self) -> AsyncRecordsLoader:
        """Access to :class:`AsyncRecords`-based read methods."""
        return self._records

    # Builder methods that configure the underlying AsyncDataLoader
    def stream(self, enabled: bool = True) -> "AsyncReadAccessor":
        """Enable or disable streaming mode (chunked reading for large files)."""
        self._loader.stream(enabled)
        return self

    def schema(self, schema: Sequence[ColumnDef]) -> "AsyncReadAccessor":
        """Set an explicit schema for the data source."""
        self._loader.schema(schema)
        return self

    def option(self, key: str, value: object) -> "AsyncReadAccessor":
        """Set a read option (e.g., header=True for CSV, multiline=True for JSON)."""
        self._loader.option(key, value)
        return self

    def options(self, **options: object) -> "AsyncReadAccessor":
        """Set multiple read options at once (PySpark-compatible).

        Args:
            **options: Dictionary of option key-value pairs

        Returns:
            Self for method chaining

        Example:
            >>> df = await db.read.options(header=True, delimiter=",").csv("data.csv")
        """
        self._loader.options(**options)
        return self

    # AsyncDataFrame read methods (delegate to AsyncDataLoader)
    async def table(self, name: str) -> AsyncDataFrame:
        """Read from a database table as an AsyncDataFrame.

        Args:
            name: Name of the table to read

        Returns:
            AsyncDataFrame that can be transformed before execution

        Example:
            >>> df = await db.read.table("users")
            >>> results = await df.collect()
        """
        return await self._loader.table(name)

    async def csv(self, path: str) -> AsyncDataFrame:
        """Read a CSV file as an AsyncDataFrame.

        Args:
            path: Path to the CSV file

        Returns:
            AsyncDataFrame containing the CSV data (lazy, materialized on .collect())
        """
        return await self._loader.csv(path)

    async def json(self, path: str) -> AsyncDataFrame:
        """Read a JSON file (array of objects) as an AsyncDataFrame.

        Args:
            path: Path to the JSON file

        Returns:
            AsyncDataFrame containing the JSON data (lazy, materialized on .collect())
        """
        return await self._loader.json(path)

    async def jsonl(self, path: str) -> AsyncDataFrame:
        """Read a JSONL file (one JSON object per line) as an AsyncDataFrame.

        Args:
            path: Path to the JSONL file

        Returns:
            AsyncDataFrame containing the JSONL data (lazy, materialized on .collect())
        """
        return await self._loader.jsonl(path)

    async def parquet(self, path: str) -> AsyncDataFrame:
        """Read a Parquet file as an AsyncDataFrame.

        Args:
            path: Path to the Parquet file

        Returns:
            AsyncDataFrame containing the Parquet data (lazy, materialized on .collect())

        Raises:
            RuntimeError: If pandas or pyarrow are not installed
        """
        return await self._loader.parquet(path)

    async def text(self, path: str, column_name: str = "value") -> AsyncDataFrame:
        """Read a text file as a single column (one line per row) as an AsyncDataFrame.

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncDataFrame containing the text file lines (lazy, materialized on .collect())
        """
        return await self._loader.text(path, column_name)

    async def textFile(self, path: str, column_name: str = "value") -> AsyncDataFrame:
        """Read a text file as a single column (PySpark-compatible alias for text()).

        Args:
            path: Path to the text file
            column_name: Name of the column to create (default: "value")

        Returns:
            AsyncDataFrame containing the text file lines (lazy, materialized on .collect())
        """
        return await self._loader.textFile(path, column_name)

    async def format(self, source: str) -> AsyncFormatReader:
        """Specify the data source format.

        Args:
            source: Format name (e.g., "csv", "json", "parquet")

        Returns:
            AsyncFormatReader for the specified format
        """
        return await self._loader.format(source)



"""SQLModel and Pydantic integration utilities for Moltres."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Type

if TYPE_CHECKING:
    pass  # Type hints only


def is_sqlmodel_model(obj: Any) -> bool:
    """Detect if an object is a SQLModel model class.

    Args:
        obj: Object to check

    Returns:
        True if obj is a SQLModel model class, False otherwise
    """
    try:
        from sqlmodel import SQLModel

        # Check if it's a class
        if not isinstance(obj, type):
            return False

        # Check if it's a subclass of SQLModel
        if issubclass(obj, SQLModel):
            # Check if it has table=True (SQLModel table models)
            # SQLModel models with table=True have __table__ attribute
            if hasattr(obj, "__table__"):
                return True
            # Also check for __tablename__ which SQLModel models can have
            if hasattr(obj, "__tablename__"):
                return True

        return False
    except ImportError:
        return False


def is_pydantic_model(obj: Any) -> bool:
    """Detect if an object is a Pydantic BaseModel class.

    Args:
        obj: Object to check

    Returns:
        True if obj is a Pydantic BaseModel class, False otherwise
    """
    try:
        from pydantic import BaseModel

        # Check if it's a class
        if not isinstance(obj, type):
            return False

        # Check if it's a subclass of BaseModel
        # Exclude SQLModel since that's handled separately
        if issubclass(obj, BaseModel):
            # Don't treat SQLModel as pure Pydantic
            try:
                from sqlmodel import SQLModel

                if issubclass(obj, SQLModel):
                    return False
            except ImportError:
                pass
            return True

        return False
    except ImportError:
        return False


def is_model_class(obj: Any) -> bool:
    """Detect if an object is a SQLModel or Pydantic model class.

    Args:
        obj: Object to check

    Returns:
        True if obj is a SQLModel or Pydantic model class, False otherwise
    """
    return is_sqlmodel_model(obj) or is_pydantic_model(obj)


def get_sqlmodel_table_name(model_class: Type) -> str:
    """Extract table name from a SQLModel model class.

    Args:
        model_class: SQLModel model class

    Returns:
        Table name

    Raises:
        ValueError: If model doesn't have a table name
    """
    if hasattr(model_class, "__tablename__"):
        tablename = getattr(model_class, "__tablename__")
        if isinstance(tablename, str):
            return tablename
    if hasattr(model_class, "__table__"):
        # Fallback to __table__.name
        table = getattr(model_class, "__table__")
        if table is not None and hasattr(table, "name"):
            return str(table.name)
    # Fallback to class name (lowercased)
    return model_class.__name__.lower()


def model_to_dict(instance: Any) -> Dict[str, Any]:
    """Convert a SQLModel or Pydantic instance to a dictionary.

    Args:
        instance: SQLModel or Pydantic instance

    Returns:
        Dictionary representation of the instance

    Raises:
        TypeError: If instance is not a SQLModel or Pydantic instance
    """
    # Try SQLModel first
    try:
        from sqlmodel import SQLModel

        if isinstance(instance, SQLModel):
            # SQLModel instances have model_dump() method (Pydantic v2) or dict() method (Pydantic v1)
            if hasattr(instance, "model_dump"):
                return instance.model_dump()
            elif hasattr(instance, "dict"):
                return instance.dict()
            else:
                # Fallback to __dict__
                return {k: v for k, v in instance.__dict__.items() if not k.startswith("_")}
    except ImportError:
        pass

    # Try Pydantic
    try:
        from pydantic import BaseModel

        if isinstance(instance, BaseModel):
            # Pydantic instances have model_dump() method (Pydantic v2) or dict() method (Pydantic v1)
            if hasattr(instance, "model_dump"):
                return instance.model_dump()
            elif hasattr(instance, "dict"):
                return instance.dict()
            else:
                # Fallback to __dict__
                return {k: v for k, v in instance.__dict__.items() if not k.startswith("_")}
    except ImportError:
        pass

    raise TypeError(f"Expected SQLModel or Pydantic instance, got {type(instance)}")


def sqlmodel_to_dict(instance: Any) -> Dict[str, Any]:
    """Convert a SQLModel instance to a dictionary.

    Args:
        instance: SQLModel instance

    Returns:
        Dictionary representation of the instance

    Raises:
        TypeError: If instance is not a SQLModel instance
    """
    return model_to_dict(instance)


def dict_to_model(data: Dict[str, Any], model_class: Type) -> Any:
    """Convert a dictionary to a SQLModel or Pydantic instance.

    Args:
        data: Dictionary containing model data
        model_class: SQLModel or Pydantic model class

    Returns:
        SQLModel or Pydantic instance

    Raises:
        TypeError: If model_class is not a SQLModel or Pydantic class
        ImportError: If required dependencies are not installed
    """
    if is_sqlmodel_model(model_class):
        # SQLModel classes can be instantiated directly with dict unpacking
        return model_class(**data)
    elif is_pydantic_model(model_class):
        # Pydantic classes can be instantiated directly with dict unpacking
        return model_class(**data)
    else:
        raise TypeError(f"Expected SQLModel or Pydantic class, got {type(model_class)}")


def dict_to_sqlmodel(data: Dict[str, Any], model_class: Type) -> Any:
    """Convert a dictionary to a SQLModel instance.

    Args:
        data: Dictionary containing model data
        model_class: SQLModel model class

    Returns:
        SQLModel instance

    Raises:
        TypeError: If model_class is not a SQLModel class
        ImportError: If SQLModel is not installed
    """
    if not is_sqlmodel_model(model_class):
        raise TypeError(f"Expected SQLModel class, got {type(model_class)}")
    return dict_to_model(data, model_class)


def rows_to_models(rows: List[Dict[str, Any]], model_class: Type) -> List[Any]:
    """Convert a list of dictionaries to SQLModel or Pydantic instances.

    Args:
        rows: List of dictionaries representing rows
        model_class: SQLModel or Pydantic model class

    Returns:
        List of SQLModel or Pydantic instances

    Raises:
        TypeError: If model_class is not a SQLModel or Pydantic class
        ImportError: If required dependencies are not installed
    """
    if not is_model_class(model_class):
        raise TypeError(f"Expected SQLModel or Pydantic class, got {type(model_class)}")

    return [dict_to_model(row, model_class) for row in rows]


def rows_to_sqlmodels(rows: List[Dict[str, Any]], model_class: Type) -> List[Any]:
    """Convert a list of dictionaries to SQLModel instances.

    Args:
        rows: List of dictionaries representing rows
        model_class: SQLModel model class

    Returns:
        List of SQLModel instances

    Raises:
        TypeError: If model_class is not a SQLModel class
        ImportError: If SQLModel is not installed
    """
    if not is_sqlmodel_model(model_class):
        raise TypeError(f"Expected SQLModel class, got {type(model_class)}")
    return rows_to_models(rows, model_class)



"""Ephemeral table management.

This module handles creation and cleanup of ephemeral (temporary) tables.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Optional, Sequence, Union

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl
    from ..dataframe.core.dataframe import DataFrame
    from ..io.records import LazyRecords, Records
    from .schema import ColumnDef
    from .table import Database

logger = logging.getLogger(__name__)


class EphemeralTableManager:
    """Handles ephemeral table creation and cleanup for Database."""

    def __init__(self, database: "Database"):
        """Initialize ephemeral table manager with a Database.

        Args:
            database: The Database instance to manage ephemeral tables for
        """
        self._db = database

    def cleanup_ephemeral_tables(self) -> None:
        """Clean up all ephemeral tables.

        Drops all tables that were marked as ephemeral.
        This is called automatically when the Database is closed.
        """
        if not self._db._ephemeral_tables:
            return
        for table_name in list(self._db._ephemeral_tables):
            try:
                self._db.drop_table(table_name, if_exists=True).collect()
            except Exception as exc:  # pragma: no cover - best effort
                logger.debug("Failed to drop ephemeral table %s: %s", table_name, exc)
        self._db._ephemeral_tables.clear()

    def create_dataframe(
        self,
        data: Union[
            Sequence[dict[str, object]],
            Sequence[tuple],
            "Records",
            "LazyRecords",
            "pd.DataFrame",
            "pl.DataFrame",
            "pl.LazyFrame",
        ],
        schema: Optional[Sequence["ColumnDef"]] = None,
        pk: Optional[Union[str, Sequence[str]]] = None,
        auto_pk: Optional[Union[str, Sequence[str]]] = None,
    ) -> "DataFrame":
        """Create a DataFrame from Python data.

        Creates a temporary table, inserts the data, and returns a DataFrame querying from that table.
        If LazyRecords is provided, it will be auto-materialized.
        If pandas/polars DataFrame or LazyFrame is provided, it will be converted to Records with lazy conversion.

        Args:
            data: Input data in one of supported formats:
                - List of dicts: [{"col1": val1, "col2": val2}, ...]
                - List of tuples: Requires schema parameter with column names
                - Records object: Extracts data and schema if available
                - LazyRecords object: Auto-materializes and extracts data and schema
                - pandas DataFrame: Converts to Records with schema preservation
                - polars DataFrame: Converts to Records with schema preservation
                - polars LazyFrame: Materializes and converts to Records with schema preservation
            schema: Optional explicit schema. If not provided, schema is inferred from data.
            pk: Optional column name(s) to mark as primary key. Can be a single string or sequence of strings for composite keys.
            auto_pk: Optional column name(s) to create as auto-incrementing primary key. Can specify same name as pk to make an existing column auto-incrementing.

        Returns:
            DataFrame querying from the created temporary table

        Raises:
            ValueError: If data is empty and no schema provided, or if primary key requirements are not met
            ValidationError: If list of tuples provided without schema, or other validation errors
        """
        from ..dataframe.core.create_dataframe import (
            ensure_primary_key,
            generate_unique_table_name,
            get_schema_from_records,
            normalize_data_to_rows,
        )
        from ..dataframe.core.dataframe import DataFrame
        from ..dataframe.io.readers.schema_inference import infer_schema_from_rows
        from ..io.records import (
            LazyRecords,
            Records,
            _is_pandas_dataframe,
            _is_polars_dataframe,
            _is_polars_lazyframe,
            _dataframe_to_records,
        )
        from ..utils.exceptions import ValidationError

        # Convert DataFrame to Records if needed
        if _is_pandas_dataframe(data) or _is_polars_dataframe(data) or _is_polars_lazyframe(data):
            data = _dataframe_to_records(data, database=self._db)

        # Normalize data to list of dicts
        # Handle LazyRecords by auto-materializing
        if isinstance(data, LazyRecords):
            materialized_records = data.collect()  # Auto-materialize
            rows = normalize_data_to_rows(materialized_records)
            # Use schema from Records if available and no explicit schema provided
            if schema is None:
                schema = get_schema_from_records(materialized_records)
        elif isinstance(data, Records):
            rows = normalize_data_to_rows(data)
            # Use schema from Records if available and no explicit schema provided
            if schema is None:
                schema = get_schema_from_records(data)
        elif isinstance(data, list) and data and isinstance(data[0], tuple):
            # Handle list of tuples - requires schema
            if schema is None:
                raise ValidationError(
                    "Schema is required when providing list of tuples. "
                    "Provide schema parameter with column names."
                )
            # Convert tuples to dicts using schema
            rows = [{col.name: val for col, val in zip(schema, row)} for row in data]
        elif isinstance(data, list):
            # List of dicts
            rows = data
        else:
            raise ValidationError(
                f"Unsupported data type: {type(data).__name__}. "
                "Expected list of dicts, list of tuples (with schema), Records, LazyRecords, "
                "pandas DataFrame, or polars DataFrame/LazyFrame."
            )

        # Validate that we have data or schema
        if not rows and not schema:
            raise ValueError("Cannot create DataFrame from empty data")

        # Infer schema if not provided
        if schema is None:
            schema = infer_schema_from_rows(rows)

        # Ensure primary key
        inferred_schema_list, new_auto_increment_cols = ensure_primary_key(
            list(schema),
            pk=pk,
            auto_pk=auto_pk,
            dialect_name=self._db._dialect_name,
            require_primary_key=False,
        )

        # Generate unique table name
        table_name = generate_unique_table_name()

        # Always use persistent staging tables so later operations (which may run on a different
        # pooled connection) can still access the data. Ephemeral cleanup happens via close().
        use_temp_tables = False
        table_handle = self._db.create_table(
            table_name,
            inferred_schema_list,
            temporary=use_temp_tables,
            if_not_exists=True,
        ).collect()
        if not use_temp_tables:
            self._db._register_ephemeral_table(table_name)

        # Insert data (exclude new auto-increment columns from INSERT)
        if rows:
            # Filter rows to only include columns that exist in schema and are not new auto-increment columns
            filtered_rows = []
            for row in rows:
                filtered_row = {
                    k: v
                    for k, v in row.items()
                    if k not in new_auto_increment_cols
                    and any(col.name == k for col in inferred_schema_list)
                }
                filtered_rows.append(filtered_row)

            records_to_insert = Records(_data=filtered_rows, _database=self._db)
            records_to_insert.insert_into(table_handle)

        # Return DataFrame querying from the temporary table
        return DataFrame.from_table(table_handle)



"""Compilation of string expressions.

This module handles compilation of string manipulation functions like concat, upper, lower, substring, etc.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from sqlalchemy import func

if TYPE_CHECKING:
    from sqlalchemy.sql import ColumnElement
    from ..expression_compiler import ExpressionCompiler


def compile_string_operation(
    compiler: "ExpressionCompiler", op: str, expression: Any
) -> "ColumnElement[Any] | None":
    """Compile a string operation expression.

    Args:
        compiler: The ExpressionCompiler instance
        op: Operation name (e.g., "concat", "upper", "lower")
        expression: :class:`Column` expression to compile

    Returns:
        Compiled SQLAlchemy column element, or None if not handled
    """
    result: "ColumnElement[Any]"
    if op == "coalesce":
        args = [compiler._compile(arg) for arg in expression.args]
        result = func.coalesce(*args)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "concat":
        args = [compiler._compile(arg) for arg in expression.args]
        # SQLite doesn't have concat() function, uses || operator instead
        if compiler.dialect.name == "sqlite":
            # Build concatenation using || operator
            result = args[0]
            for arg in args[1:]:
                result = result.op("||")(arg)
        else:
            # PostgreSQL and MySQL support concat() function
            result = func.concat(*args)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "upper":
        result = func.upper(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "lower":
        result = func.lower(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "substring":
        col_expr = compiler._compile(expression.args[0])
        pos = expression.args[1]
        if len(expression.args) > 2:
            length = expression.args[2]
            result = func.substring(col_expr, pos, length)
        else:
            result = func.substring(col_expr, pos)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "trim":
        result = func.trim(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "ltrim":
        result = func.ltrim(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "rtrim":
        result = func.rtrim(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "initcap":
        col_expr = compiler._compile(expression.args[0])
        # Use dialect-specific initcap
        if compiler.dialect.name == "postgresql":
            result = func.initcap(col_expr)
        elif compiler.dialect.name == "duckdb":
            from ...utils.exceptions import CompilationError

            raise CompilationError(
                f"initcap() is not supported for {compiler.dialect.name} dialect. "
                "Supported dialects: PostgreSQL"
            )
        else:
            # MySQL/SQLite: not directly supported, use literal_column for workaround
            from sqlalchemy import literal_column

            result = literal_column(f"INITCAP({col_expr})")
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "instr":
        col_expr = compiler._compile(expression.args[0])
        substr_expr = compiler._compile(expression.args[1])
        # Use dialect-specific instr
        if compiler.dialect.name == "postgresql":
            result = func.strpos(col_expr, substr_expr)
        elif compiler.dialect.name == "mysql":
            result = func.locate(substr_expr, col_expr)
        else:
            # SQLite: instr
            result = func.instr(col_expr, substr_expr)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "locate":
        substr_expr = compiler._compile(expression.args[0])
        col_expr = compiler._compile(expression.args[1])
        pos = expression.args[2] if len(expression.args) > 2 else 1
        from sqlalchemy import literal

        # Use dialect-specific locate
        if compiler.dialect.name == "postgresql":
            # PostgreSQL: strpos doesn't support start position, use substring
            if pos > 1:
                from sqlalchemy import literal_column

                result = func.strpos(func.substring(col_expr, pos), substr_expr) + literal(pos - 1)
            else:
                result = func.strpos(col_expr, substr_expr)
        elif compiler.dialect.name == "mysql":
            result = func.locate(substr_expr, col_expr, pos)
        else:
            # SQLite: instr with offset
            if pos > 1:
                from sqlalchemy import literal_column

                result = func.instr(func.substring(col_expr, pos), substr_expr) + literal(pos - 1)
            else:
                result = func.instr(col_expr, substr_expr)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "translate":
        col_expr = compiler._compile(expression.args[0])
        from_chars = expression.args[1]
        to_chars = expression.args[2]
        # Use dialect-specific translate
        if compiler.dialect.name == "postgresql":
            result = func.translate(col_expr, from_chars, to_chars)
        elif compiler.dialect.name == "duckdb":
            from ...utils.exceptions import CompilationError

            raise CompilationError(
                f"translate() is not supported for {compiler.dialect.name} dialect. "
                "Supported dialects: PostgreSQL"
            )
        else:
            # MySQL/SQLite: requires workaround (not directly supported)
            from ...utils.exceptions import CompilationError

            raise CompilationError(
                f"translate() is not supported for {compiler.dialect.name} dialect. "
                "PostgreSQL supports this function natively."
            )
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "regexp_extract":
        # SQLAlchemy doesn't have a direct regexp_extract, use dialect-specific function
        col_expr = compiler._compile(expression.args[0])
        pattern = expression.args[1]
        group_idx = expression.args[2] if len(expression.args) > 2 else 0
        # Use func for dialect-specific regex functions
        result = func.regexp_extract(col_expr, pattern, group_idx)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "regexp_replace":
        col_expr = compiler._compile(expression.args[0])
        pattern = expression.args[1]
        replacement = expression.args[2]
        result = func.regexp_replace(col_expr, pattern, replacement)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "split":
        # SQLAlchemy doesn't have split, use string_to_array or similar
        col_expr = compiler._compile(expression.args[0])
        delimiter = expression.args[1]
        result = func.string_to_array(col_expr, delimiter)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "replace":
        col_expr = compiler._compile(expression.args[0])
        search = expression.args[1]
        replacement = expression.args[2]
        result = func.replace(col_expr, search, replacement)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "length":
        result = func.length(compiler._compile(expression.args[0]))
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "lpad":
        col_expr = compiler._compile(expression.args[0])
        length = expression.args[1]
        pad = expression.args[2] if len(expression.args) > 2 else " "
        result = func.lpad(col_expr, length, pad)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "rpad":
        col_expr = compiler._compile(expression.args[0])
        length = expression.args[1]
        pad = expression.args[2] if len(expression.args) > 2 else " "
        result = func.rpad(col_expr, length, pad)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "greatest":
        args = [compiler._compile(arg) for arg in expression.args]
        result = func.greatest(*args)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    if op == "least":
        args = [compiler._compile(arg) for arg in expression.args]
        result = func.least(*args)
        if expression._alias:
            result = result.label(expression._alias)
        return result

    return None  # Not handled by this module



"""DataFrame interface implementations (pandas, polars, async variants)."""

from .pandas_dataframe import PandasDataFrame
from .polars_dataframe import PolarsDataFrame
from .async_pandas_dataframe import AsyncPandasDataFrame
from .async_polars_dataframe import AsyncPolarsDataFrame

__all__ = [
    "PandasDataFrame",
    "PolarsDataFrame",
    "AsyncPandasDataFrame",
    "AsyncPolarsDataFrame",
]



"""Helper utilities for SQL generation."""

from __future__ import annotations

import re
from collections.abc import Iterable

from ..utils.exceptions import ValidationError


def comma_separated(values: Iterable[str]) -> str:
    return ", ".join(values)


def quote_identifier(identifier: str, quote_char: str = '"') -> str:
    """Quote a SQL identifier, validating it for safety.

    Args:
        identifier: The identifier to quote (e.g., "table_name" or "schema.table")
        quote_char: The character to use for quoting (default: double quote)

    Returns:
        The quoted identifier

    Raises:
        ValidationError: If the identifier is empty or contains invalid characters
    """
    if not identifier or not identifier.strip():
        raise ValidationError("SQL identifier cannot be empty")

    # Validate identifier parts (alphanumeric, underscore, and dot for qualified names)
    # SQL identifiers can contain letters, digits, underscores, and dots for schema.table
    # We allow dots for qualified names like "schema.table"
    parts = identifier.split(".")
    for part in parts:
        if not part:
            raise ValidationError(f"SQL identifier contains empty part: {identifier!r}")
        # Check for SQL injection patterns (semicolons, comments, etc.)
        if re.search(r"[;\'\"\\]", part):
            raise ValidationError(
                f"SQL identifier contains invalid characters: {identifier!r}. "
                "Identifiers may only contain letters, digits, underscores, and dots."
            )

    quoted = [f"{quote_char}{part}{quote_char}" for part in parts if part]
    return ".".join(quoted) if quoted else identifier


def format_literal(value: object) -> str:
    if value is None:
        return "NULL"
    if isinstance(value, bool):
        return "TRUE" if value else "FALSE"
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, str):
        escaped = value.replace("'", "''")
        return f"'{escaped}'"
    raise TypeError(f"Unsupported literal type: {type(value)!r}")



"""Format-specific readers for :class:`DataFrame` operations."""

from __future__ import annotations

from .csv_reader import read_csv, read_csv_stream
from .json_reader import read_json, read_json_stream, read_jsonl, read_jsonl_stream
from .parquet_reader import read_parquet, read_parquet_stream
from .schema_inference import apply_schema_to_rows, infer_schema_from_rows
from .text_reader import read_text, read_text_stream

__all__ = [
    "apply_schema_to_rows",
    "infer_schema_from_rows",
    "read_csv",
    "read_csv_stream",
    "read_json",
    "read_json_stream",
    "read_jsonl",
    "read_jsonl_stream",
    "read_parquet",
    "read_parquet_stream",
    "read_text",
    "read_text_stream",
]



""":class:`Column` wrapper for pandas-style string accessor support."""

from __future__ import annotations

from ...expressions.column import Column
from .base_column_wrapper import BaseColumnWrapper


class PandasColumn(BaseColumnWrapper):
    """Wrapper around :class:`Column` that adds pandas-style string accessor.

    This class wraps a :class:`Column` expression and adds a `str` attribute
    that provides string operations like pandas :class:`DataFrame`.str.

    Example:
        >>> col = PandasColumn(col("name"))
        >>> col.str.upper()  # Returns :class:`Column` expression for UPPER(name)
    """

    def __init__(self, column: Column):
        """Initialize with a :class:`Column` expression.

        Args:
            column: The :class:`Column` expression to wrap
        """
        super().__init__(column)

        # Add str accessor if available
        try:
            from .pandas_string_accessor import _StringAccessor

            self.str = _StringAccessor(column)
        except ImportError:
            self.str = None  # type: ignore



"""Factory helpers for logical plan nodes."""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Dict, Optional

from ..expressions.column import Column

if TYPE_CHECKING:
    from ..table.schema import ColumnDef
from .plan import (
    Aggregate,
    AntiJoin,
    CTE,
    Distinct,
    Except,
    Explode,
    FileScan,
    Filter,
    GroupedPivot,
    Intersect,
    Join,
    Limit,
    LogicalPlan,
    Pivot,
    Project,
    RawSQL,
    RecursiveCTE,
    Sample,
    SemiJoin,
    Sort,
    SortOrder,
    TableScan,
    Union,
)


def scan(table: str, alias: str | None = None) -> TableScan:
    """Create a TableScan logical plan node.

    Args:
        table: Name of the table to scan
        alias: Optional alias for the table

    Returns:
        TableScan logical plan node
    """
    return TableScan(table=table, alias=alias)


def file_scan(
    path: str,
    format: str,
    schema: Optional[Sequence["ColumnDef"]] = None,
    options: Optional[Dict[str, object]] = None,
    column_name: Optional[str] = None,
) -> FileScan:
    """Create a FileScan logical plan node.

    Args:
        path: Path to the file
        format: File format ("csv", "json", "jsonl", "parquet", "text")
        schema: Optional explicit schema for the file data
        options: Dictionary of format-specific read options
        column_name: :class:`Column` name for text files (default: "value")

    Returns:
        FileScan logical plan node
    """
    return FileScan(
        path=path,
        format=format,
        schema=schema,
        options=options or {},
        column_name=column_name,
    )


def project(child: LogicalPlan, columns: Sequence[Column]) -> Project:
    """Create a Project logical plan node.

    Args:
        child: Child logical plan
        columns: Sequence of column expressions to project

    Returns:
        Project logical plan node
    """
    return Project(child=child, projections=tuple(columns))


def filter(child: LogicalPlan, predicate: Column) -> Filter:
    """Create a Filter logical plan node.

    Args:
        child: Child logical plan
        predicate: :class:`Column` expression for the filter condition

    Returns:
        Filter logical plan node
    """
    return Filter(child=child, predicate=predicate)


def limit(child: LogicalPlan, count: int, offset: int = 0) -> Limit:
    """Create a Limit logical plan node.

    Args:
        child: Child logical plan
        count: Maximum number of rows to return
        offset: Number of rows to skip before returning results

    Returns:
        Limit logical plan node
    """
    return Limit(child=child, count=count, offset=offset)


def sample(child: LogicalPlan, fraction: float, seed: Optional[int] = None) -> Sample:
    """Create a Sample logical plan node.

    Args:
        child: Child logical plan
        fraction: Fraction of rows to sample (0.0 to 1.0)
        seed: Optional random seed for reproducible sampling

    Returns:
        Sample logical plan node
    """
    if not 0.0 <= fraction <= 1.0:
        raise ValueError(f"Sample fraction must be between 0.0 and 1.0, got {fraction}")
    return Sample(child=child, fraction=fraction, seed=seed)


def distinct(child: LogicalPlan) -> Distinct:
    """Create a Distinct logical plan node.

    Args:
        child: Child logical plan

    Returns:
        Distinct logical plan node
    """
    return Distinct(child=child)


def union(left: LogicalPlan, right: LogicalPlan, distinct: bool = True) -> Union:
    """Create a Union logical plan node.

    Args:
        left: Left logical plan
        right: Right logical plan
        distinct: If True, use UNION (distinct), if False use UNION ALL

    Returns:
        Union logical plan node
    """
    return Union(left=left, right=right, distinct=distinct)


def intersect(left: LogicalPlan, right: LogicalPlan, distinct: bool = True) -> Intersect:
    """Create an Intersect logical plan node.

    Args:
        left: Left logical plan
        right: Right logical plan
        distinct: If True, use INTERSECT (distinct), if False use INTERSECT ALL

    Returns:
        Intersect logical plan node
    """
    return Intersect(left=left, right=right, distinct=distinct)


def except_(left: LogicalPlan, right: LogicalPlan, distinct: bool = True) -> Except:
    """Create an Except logical plan node.

    Args:
        left: Left logical plan
        right: Right logical plan
        distinct: If True, use EXCEPT (distinct), if False use EXCEPT ALL

    Returns:
        Except logical plan node
    """
    return Except(left=left, right=right, distinct=distinct)


def order_by(child: LogicalPlan, orders: Iterable[SortOrder]) -> Sort:
    """Create a Sort logical plan node.

    Args:
        child: Child logical plan
        orders: Iterable of SortOrder objects defining sort criteria

    Returns:
        Sort logical plan node
    """
    return Sort(child=child, orders=tuple(orders))


def sort_order(expression: Column, descending: bool = False) -> SortOrder:
    """Create a SortOrder specification.

    Args:
        expression: :class:`Column` expression to sort by
        descending: If True, sort in descending order (default: False)

    Returns:
        SortOrder specification
    """
    return SortOrder(expression=expression, descending=descending)


def aggregate(
    child: LogicalPlan, keys: Sequence[Column], aggregates: Sequence[Column]
) -> Aggregate:
    """Create an Aggregate logical plan node.

    Args:
        child: Child logical plan
        keys: Sequence of column expressions for grouping
        aggregates: Sequence of aggregate column expressions

    Returns:
        Aggregate logical plan node
    """
    return Aggregate(child=child, grouping=tuple(keys), aggregates=tuple(aggregates))


def join(
    left: LogicalPlan,
    right: LogicalPlan,
    *,
    how: str,
    on: Sequence[tuple[str, str]] | None = None,
    condition: Column | None = None,
    lateral: bool = False,
    hints: Sequence[str] | None = None,
) -> Join:
    """Create a Join logical plan node.

    Args:
        left: Left logical plan
        right: Right logical plan
        how: Join type ("inner", "left", "right", "full", "cross")
        on: Optional sequence of (left_column, right_column) tuples for equality joins
        condition: Optional column expression for custom join condition
        lateral: If True, create a LATERAL join (PostgreSQL, MySQL 8.0+)

    Returns:
        Join logical plan node

    Raises:
        CompilationError: If neither 'on' nor 'condition' is provided
    """
    hints_tuple = tuple(hints) if hints else None
    return Join(
        left=left,
        right=right,
        how=how,
        on=None if on is None else tuple(on),
        condition=condition,
        lateral=lateral,
        hints=hints_tuple,
    )


def cte(plan: LogicalPlan, name: str) -> CTE:
    """Create a CTE (Common Table Expression) logical plan node.

    Args:
        plan: Logical plan to wrap as a CTE
        name: Name for the CTE

    Returns:
        CTE logical plan node
    """
    return CTE(name=name, child=plan)


def recursive_cte(
    name: str, initial: LogicalPlan, recursive: LogicalPlan, union_all: bool = False
) -> RecursiveCTE:
    """Create a RecursiveCTE (WITH RECURSIVE) logical plan node.

    Args:
        name: Name for the recursive CTE
        initial: Initial/seed query (non-recursive part)
        recursive: Recursive part that references the CTE
        union_all: If True, use UNION ALL; if False, use UNION (distinct)

    Returns:
        RecursiveCTE logical plan node

    Example:
        >>> # Fibonacci sequence
            >>> from moltres.expressions import functions as F
            >>> initial = scan("seed").select(F.lit(1).alias("n"), F.lit(1).alias("fib"))
        >>> recursive = scan(name).select(...)  # References CTE name
        >>> fib_cte = recursive_cte("fib", initial, recursive)
    """
    return RecursiveCTE(name=name, initial=initial, recursive=recursive, union_all=union_all)


def semi_join(
    left: LogicalPlan,
    right: LogicalPlan,
    *,
    on: Sequence[tuple[str, str]] | None = None,
    condition: Column | None = None,
) -> SemiJoin:
    """Create a SemiJoin logical plan node (EXISTS subquery).

    Args:
        left: Left logical plan
        right: Right logical plan (subquery for EXISTS)
        on: Optional sequence of (left_column, right_column) tuples for equality joins
        condition: Optional column expression for custom join condition

    Returns:
        SemiJoin logical plan node
    """
    return SemiJoin(
        left=left, right=right, on=None if on is None else tuple(on), condition=condition
    )


def anti_join(
    left: LogicalPlan,
    right: LogicalPlan,
    *,
    on: Sequence[tuple[str, str]] | None = None,
    condition: Column | None = None,
) -> AntiJoin:
    """Create an AntiJoin logical plan node (NOT EXISTS subquery).

    Args:
        left: Left logical plan
        right: Right logical plan (subquery for NOT EXISTS)
        on: Optional sequence of (left_column, right_column) tuples for equality joins
        condition: Optional column expression for custom join condition

    Returns:
        AntiJoin logical plan node
    """
    return AntiJoin(
        left=left, right=right, on=None if on is None else tuple(on), condition=condition
    )


def pivot(
    child: LogicalPlan,
    pivot_column: str,
    value_column: str,
    agg_func: str = "sum",
    pivot_values: Sequence[str] | None = None,
) -> Pivot:
    """Create a Pivot logical plan node.

    Args:
        child: Child logical plan
        pivot_column: :class:`Column` to pivot on (becomes column headers)
        value_column: :class:`Column` containing values to aggregate
        agg_func: Aggregation function name (e.g., "sum", "avg", "count")
        pivot_values: Optional list of specific values to pivot (if None, uses all distinct values)

    Returns:
        Pivot logical plan node
    """
    pivot_values_tuple = tuple(pivot_values) if pivot_values else None
    return Pivot(
        child=child,
        pivot_column=pivot_column,
        value_column=value_column,
        agg_func=agg_func,
        pivot_values=pivot_values_tuple,
    )


def grouped_pivot(
    child: LogicalPlan,
    grouping: tuple[Column, ...],
    pivot_column: str,
    value_column: str,
    agg_func: str = "sum",
    pivot_values: Sequence[str] | None = None,
) -> GroupedPivot:
    """Create a GroupedPivot logical plan node.

    Args:
        child: Child logical plan
        grouping: Columns to group by
        pivot_column: :class:`Column` to pivot on (becomes column headers)
        value_column: :class:`Column` containing values to aggregate
        agg_func: Aggregation function name (e.g., "sum", "avg", "count")
        pivot_values: Optional list of specific values to pivot (if None, must be provided or discovered)

    Returns:
        GroupedPivot logical plan node
    """
    pivot_values_tuple = tuple(pivot_values) if pivot_values else None
    return GroupedPivot(
        child=child,
        grouping=grouping,
        pivot_column=pivot_column,
        value_column=value_column,
        agg_func=agg_func,
        pivot_values=pivot_values_tuple,
    )


def explode(child: LogicalPlan, column: Column, alias: str = "value") -> Explode:
    """Create an Explode logical plan node (expands array/JSON column into multiple rows).

    Args:
        child: Child logical plan
        column: :class:`Column` expression to explode (array or JSON column)
        alias: Alias for the exploded value column (default: "value")

    Returns:
        Explode logical plan node
    """
    return Explode(child=child, column=column, alias=alias)


def raw_sql(sql: str, params: Dict[str, object] | None = None) -> RawSQL:
    """Create a RawSQL logical plan node.

    Args:
        sql: SQL query string to execute
        params: Optional dictionary of named parameters for parameterized queries

    Returns:
        RawSQL logical plan node
    """
    return RawSQL(sql=sql, params=params)



"""Grouped :class:`DataFrame` helper."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, Optional, Sequence, Union

from ...expressions.column import Column, col
from ...logical import operators
from ...logical.plan import LogicalPlan
from ..core.dataframe import DataFrame


@dataclass(frozen=True)
class GroupedDataFrame:
    """Represents a :class:`DataFrame` grouped by one or more columns.

    This is returned by :class:`DataFrame`.group_by() and provides aggregation methods.
    """

    plan: LogicalPlan
    keys: tuple[Column, ...]
    parent: DataFrame

    def agg(self, *aggregations: Union[Column, str, Dict[str, str]]) -> DataFrame:
        """Apply aggregation functions to the grouped data.

        Args:
            *aggregations: One or more aggregation expressions. Can be:
                - :class:`Column` expressions (e.g., sum(col("amount")))
                - String column names (e.g., "amount" - defaults to sum())
                - Dictionary mapping column names to aggregation functions
                  (e.g., {"amount": "sum", "price": "avg"})

        Returns:
            :class:`DataFrame` with aggregated results

        Raises:
            ValueError: If no aggregations are provided or if invalid
                aggregation expressions are used

        Example:
            >>> from moltres import connect, col
            >>> from moltres.expressions import functions as F
            >>> from moltres.table.schema import column
            >>> db = connect("sqlite:///:memory:")
            >>> db.create_table("sales", [column("category", "TEXT"), column("amount", "REAL"), column("price", "REAL")]).collect()
            >>> from moltres.io.records import :class:`Records`
            >>> :class:`Records`(_data=[{"category": "A", "amount": 100.0, "price": 10.0}, {"category": "A", "amount": 200.0, "price": 20.0}, {"category": "B", "amount": 150.0, "price": 15.0}], _database=db).insert_into("sales")
            >>> # Using :class:`Column` expressions
            >>> df = db.table("sales").select()
            >>> result = df.group_by("category").agg(F.sum(col("amount")).alias("total"), F.avg(col("price")).alias("avg_price"))
            >>> results = result.collect()
            >>> len(results)
            2
            >>> results[0]["total"]
            300.0
            >>> # Using string column names (defaults to sum)
            >>> result2 = df.group_by("category").agg("amount")
            >>> results2 = result2.collect()
            >>> results2[0]["amount"]
            300.0
            >>> # Using dictionary syntax
            >>> result3 = df.group_by("category").agg({"amount": "sum", "price": "avg"})
            >>> results3 = result3.collect()
            >>> results3[0]["amount"]
            300.0
            >>> db.close()
        """
        if not aggregations:
            raise ValueError("agg requires at least one aggregation expression")

        # Normalize all aggregations to Column expressions
        from ..helpers.groupby_helpers import normalize_aggregations, validate_aggregation

        # Allow empty aggregations for special cases like dropDuplicates
        normalized_aggs = normalize_aggregations(
            aggregations, alias_with_column_name=True, allow_empty=True
        )

        # If no aggregations, just return grouping columns (for dropDuplicates)
        if not normalized_aggs:
            # Select only grouping columns and apply distinct
            grouping_cols = list(self.keys)
            plan = operators.project(self.plan, tuple(grouping_cols))
            plan = operators.distinct(plan)  # type: ignore[assignment]
        else:
            normalized = tuple(validate_aggregation(expr) for expr in normalized_aggs)
            plan = operators.aggregate(self.plan, self.keys, normalized)  # type: ignore[assignment]
        return DataFrame(plan=plan, database=self.parent.database)

    @staticmethod
    def _create_aggregation_from_string(column_name: str, func_name: str) -> Column:
        """Create an aggregation :class:`Column` from a column name and function name string.

        Args:
            column_name: Name of the column to aggregate
            func_name: Name of the aggregation function (e.g., "sum", "avg", "min", "max", "count")

        Returns:
            :class:`Column` expression for the aggregation

        Raises:
            ValueError: If the function name is not recognized
        """
        from ..helpers.groupby_helpers import create_aggregation_from_string

        return create_aggregation_from_string(column_name, func_name)

    def pivot(
        self, pivot_col: str, values: Optional[Sequence[str]] = None
    ) -> "PivotedGroupedDataFrame":
        """Pivot the grouped data on a column.

        Args:
            pivot_col: :class:`Column` to pivot on (values become column headers)
            values: Optional list of specific values to pivot (if None, must be provided later or discovered)

        Returns:
            PivotedGroupedDataFrame that can be aggregated

        Example:
            >>> df.group_by("category").pivot("status").agg("amount")
            >>> df.group_by("category").pivot("status", values=["active", "inactive"]).agg("amount")
        """
        return PivotedGroupedDataFrame(
            plan=self.plan,
            keys=self.keys,
            pivot_column=pivot_col,
            pivot_values=tuple(values) if values else None,
            parent=self.parent,
        )

    @staticmethod
    def _validate_aggregation(expr: Column) -> Column:
        """Validate that an expression is a valid aggregation.

        Args:
            expr: :class:`Column` expression to validate

        Returns:
            The validated column expression

        Raises:
            ValueError: If the expression is not a valid aggregation
        """
        if not expr.op.startswith("agg_"):
            raise ValueError(
                "Aggregation expressions must be created with moltres aggregate helpers "
                "(e.g., sum(), avg(), count(), min(), max())"
            )
        return expr


@dataclass(frozen=True)
class PivotedGroupedDataFrame:
    """Represents a :class:`DataFrame` grouped by columns with a pivot operation applied.

    This is returned by :class:`GroupedDataFrame`.pivot() and provides aggregation methods
    that will create pivoted columns.
    """

    plan: LogicalPlan
    keys: tuple[Column, ...]
    pivot_column: str
    pivot_values: Optional[tuple[str, ...]]
    parent: DataFrame

    def agg(self, *aggregations: Union[Column, str, Dict[str, str]]) -> DataFrame:
        """Apply aggregation functions to the pivoted grouped data.

        Args:
            *aggregations: One or more aggregation expressions. Can be:
                - :class:`Column` expressions (e.g., sum(col("amount")))
                - String column names (e.g., "amount" - defaults to sum())
                - Dictionary mapping column names to aggregation functions
                  (e.g., {"amount": "sum", "price": "avg"})

        Returns:
            :class:`DataFrame` with pivoted aggregated results

        Raises:
            ValueError: If no aggregations are provided or if invalid
                aggregation expressions are used

        Example:
            >>> from moltres import col
            >>> from moltres.expressions import functions as F
            >>> # Using string column name
            >>> df.group_by("category").pivot("status").agg("amount")

            >>> # Using :class:`Column` expression
            >>> df.group_by("category").pivot("status").agg(F.sum(col("amount")))

            >>> # With specific pivot values
            >>> df.group_by("category").pivot("status", values=["active", "inactive"]).agg("amount")
        """
        if not aggregations:
            raise ValueError("agg requires at least one aggregation expression")

        # Normalize all aggregations to Column expressions
        from ..helpers.groupby_helpers import (
            normalize_aggregations,
            validate_aggregation,
            extract_value_column,
            extract_agg_func,
        )

        normalized_aggs = normalize_aggregations(aggregations, alias_with_column_name=False)

        # For pivoted grouped data, we can only aggregate one column at a time
        # (PySpark behavior - pivot with multiple aggregations requires different syntax)
        if len(normalized_aggs) > 1:
            raise ValueError(
                "Pivoted grouped aggregation supports only one aggregation expression. "
                "Multiple aggregations are not supported with pivot."
            )

        agg_expr = normalized_aggs[0]
        validate_aggregation(agg_expr)

        # Extract the value column from the aggregation
        # For sum(col("amount")), we need "amount"
        value_column = extract_value_column(agg_expr)

        # Extract the aggregation function name
        agg_func = extract_agg_func(agg_expr)

        # If pivot_values is not provided, infer them from the data (PySpark behavior)
        pivot_values = self.pivot_values
        if pivot_values is None:
            # Query distinct values from the pivot column
            distinct_df = DataFrame(plan=self.plan, database=self.parent.database)
            distinct_df = distinct_df.select(col(self.pivot_column)).distinct()
            distinct_rows = distinct_df.collect()
            pivot_values = tuple(
                str(row[self.pivot_column])
                for row in distinct_rows
                if row[self.pivot_column] is not None
            )

            if not pivot_values:
                raise ValueError(
                    f"No distinct values found in pivot column '{self.pivot_column}'. "
                    "Please provide pivot_values explicitly."
                )

        # Create a GroupedPivot logical plan
        plan = operators.grouped_pivot(
            self.plan,
            grouping=self.keys,
            pivot_column=self.pivot_column,
            value_column=value_column,
            agg_func=agg_func,
            pivot_values=pivot_values,
        )
        return DataFrame(plan=plan, database=self.parent.database)

    @staticmethod
    def _extract_value_column(agg_expr: Column) -> str:
        """Extract the column name from an aggregation expression.

        Args:
            agg_expr: Aggregation :class:`Column` expression (e.g., sum(col("amount")))

        Returns:
            :class:`Column` name string (e.g., "amount")

        Raises:
            ValueError: If the column cannot be extracted
        """
        from ..helpers.groupby_helpers import extract_value_column

        return extract_value_column(agg_expr)

    @staticmethod
    def _extract_agg_func(agg_expr: Column) -> str:
        """Extract the aggregation function name from an aggregation expression.

        Args:
            agg_expr: Aggregation :class:`Column` expression (e.g., sum(col("amount")))

        Returns:
            Aggregation function name (e.g., "sum")
        """
        from ..helpers.groupby_helpers import extract_agg_func

        return extract_agg_func(agg_expr)

    @staticmethod
    def _create_aggregation_from_string(column_name: str, func_name: str) -> Column:
        """Create an aggregation :class:`Column` from a column name and function name string.

        Args:
            column_name: Name of the column to aggregate
            func_name: Name of the aggregation function (e.g., "sum", "avg", "min", "max", "count")

        Returns:
            :class:`Column` expression for the aggregation

        Raises:
            ValueError: If the function name is not recognized
        """
        from ..helpers.groupby_helpers import create_aggregation_from_string

        return create_aggregation_from_string(column_name, func_name)

    @staticmethod
    def _validate_aggregation(expr: Column) -> Column:
        """Validate that an expression is a valid aggregation.

        Args:
            expr: :class:`Column` expression to validate

        Returns:
            The validated column expression

        Raises:
            ValueError: If the expression is not a valid aggregation
        """
        from ..helpers.groupby_helpers import validate_aggregation

        return validate_aggregation(expr)
